# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for honggfuzz engine."""
# pylint: disable=unused-argument

import os
import shutil
import unittest

from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.bot.fuzzers.honggfuzz import engine
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils

TEST_PATH = os.path.abspath(os.path.dirname(__file__))
DATA_DIR = os.path.join(TEST_PATH, 'test_data')
TEMP_DIR = os.path.join(TEST_PATH, 'temp')


def clear_temp_dir():
  """Clear temp directory."""
  if os.path.exists(TEMP_DIR):
    shutil.rmtree(TEMP_DIR)

  os.mkdir(TEMP_DIR)


def setup_testcase_and_corpus(testcase, corpus):
  """Setup testcase and corpus."""
  clear_temp_dir()
  copied_testcase_path = os.path.join(TEMP_DIR, testcase)
  shutil.copy(os.path.join(DATA_DIR, testcase), copied_testcase_path)

  copied_corpus_path = os.path.join(TEMP_DIR, corpus)
  src_corpus_path = os.path.join(DATA_DIR, corpus)

  if os.path.exists(src_corpus_path):
    shutil.copytree(src_corpus_path, copied_corpus_path)
  else:
    os.mkdir(copied_corpus_path)

  return copied_testcase_path, copied_corpus_path


@test_utils.integration
class IntegrationTest(unittest.TestCase):
  """Integration tests."""

  def setUp(self):
    self.maxDiff = None  # pylint: disable=invalid-name
    test_helpers.patch_environ(self)

    os.environ['BUILD_DIR'] = DATA_DIR

  def assert_has_stats(self, results):
    """Assert that stats exist."""
    self.assertIn('iterations', results.stats)
    self.assertIn('time', results.stats)
    self.assertIn('speed', results.stats)
    self.assertIn('crashes_count', results.stats)
    self.assertIn('timeout_count', results.stats)
    self.assertIn('new_units_added', results.stats)
    self.assertIn('slowest_unit_ms', results.stats)
    self.assertIn('guard_nb', results.stats)
    self.assertIn('branch_coverage_percent', results.stats)
    self.assertIn('peak_rss_mb', results.stats)

  def test_reproduce(self):
    """Tests reproducing a crash."""
    testcase_path, _ = setup_testcase_and_corpus('crash', 'empty_corpus')
    engine_impl = engine.HonggfuzzEngine()
    target_path = engine_common.find_fuzzer_path(DATA_DIR, 'test_fuzzer')
    result = engine_impl.reproduce(target_path, testcase_path, [], 65)
    self.assertListEqual([target_path], result.command)
    self.assertIn('ERROR: AddressSanitizer: heap-use-after-free', result.output)

  @test_utils.slow
  def test_fuzz_no_crash(self):
    """Test fuzzing (no crash)."""
    _, corpus_path = setup_testcase_and_corpus('empty', 'corpus')
    engine_impl = engine.HonggfuzzEngine()
    target_path = engine_common.find_fuzzer_path(DATA_DIR, 'test_fuzzer')
    options = engine_impl.prepare(corpus_path, target_path, DATA_DIR)
    results = engine_impl.fuzz(target_path, options, TEMP_DIR, 10)
    self.assertListEqual([
        os.path.join(DATA_DIR, 'honggfuzz'),
        '-n',
        '1',
        '--exit_upon_crash',
        '-v',
        '-z',
        '-P',
        '-S',
        '--rlimit_rss',
        '2560',
        '--timeout',
        '25',
        '--dict',
        os.path.join(DATA_DIR, 'test_fuzzer.dict'),
        '--input',
        os.path.join(TEMP_DIR, 'corpus'),
        '--workspace',
        TEMP_DIR,
        '--run_time',
        '10',
        '--',
        target_path,
    ], results.command)

    self.assertGreater(len(os.listdir(corpus_path)), 0)
    self.assert_has_stats(results)

  def test_fuzz_crash(self):
    """Test fuzzing that results in a crash."""
    _, corpus_path = setup_testcase_and_corpus('empty', 'corpus')
    engine_impl = engine.HonggfuzzEngine()
    target_path = engine_common.find_fuzzer_path(DATA_DIR,
                                                 'always_crash_fuzzer')
    options = engine_impl.prepare(corpus_path, target_path, DATA_DIR)
    results = engine_impl.fuzz(target_path, options, TEMP_DIR, 10)
    self.assertListEqual([
        os.path.join(DATA_DIR, 'honggfuzz'),
        '-n',
        '1',
        '--exit_upon_crash',
        '-v',
        '-z',
        '-P',
        '-S',
        '--rlimit_rss',
        '2560',
        '--timeout',
        '25',
        '--input',
        os.path.join(TEMP_DIR, 'corpus'),
        '--workspace',
        TEMP_DIR,
        '--run_time',
        '10',
        '--',
        target_path,
    ], results.command)

    self.assertIn('Seen a crash. Terminating all fuzzing threads', results.logs)
    self.assertEqual(1, len(results.crashes))
    crash = results.crashes[0]
    self.assertEqual(TEMP_DIR, os.path.dirname(crash.input_path))
    self.assertIn('ERROR: AddressSanitizer: heap-use-after-free',
                  crash.stacktrace)

    with open(crash.input_path, 'rb') as f:
      self.assertEqual(b'A', f.read()[:1])

    self.assert_has_stats(results)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Runs http(s) server in the background."""

import http.server
import mimetypes
import os
import socket
import threading

from clusterfuzz._internal.system import environment


def get_absolute_testcase_file(request_path):
  """Search the input directory and additional paths for the requested file."""
  # Gather the list of search path directories.
  current_working_directory = os.getcwd()
  data_directory = environment.get_value('FUZZ_DATA')
  input_directory = environment.get_value('INPUT_DIR')
  fuzzer_directory = environment.get_value('FUZZERS_DIR')
  layout_tests_directory = os.path.join(data_directory, 'LayoutTests')
  layout_tests_http_tests_directory = os.path.join(layout_tests_directory,
                                                   'http', 'tests')
  layout_tests_wpt_tests_directory = os.path.join(layout_tests_directory,
                                                  'external', 'wpt')

  # TODO(mbarbella): Add support for aliasing and directories from
  # https://cs.chromium.org/chromium/src/third_party/blink/tools/blinkpy/web_tests/servers/apache_http.py?q=apache_http.py&sq=package:chromium&dr&l=60

  # Check all search paths for the requested file.
  search_paths = [
      current_working_directory,
      fuzzer_directory,
      input_directory,
      layout_tests_directory,
      layout_tests_http_tests_directory,
      layout_tests_wpt_tests_directory,
  ]
  for search_path in search_paths:
    base_string = search_path + os.path.sep
    path = request_path.lstrip('/')
    if not path or path.endswith('/'):
      path += 'index.html'
    absolute_path = os.path.abspath(os.path.join(search_path, path))
    if (absolute_path.startswith(base_string) and
        os.path.exists(absolute_path) and not os.path.isdir(absolute_path)):
      return absolute_path

  return None


def guess_mime_type(filename):
  """Guess mime type based of file extension."""
  if not mimetypes.inited:
    mimetypes.init()

  return mimetypes.guess_type(filename)[0]


class BotHTTPServer(http.server.HTTPServer):
  """Host the bot's test case directories over HTTP."""

  def __init__(self, server_address, handler_class):
    http.server.HTTPServer.__init__(self, server_address, handler_class)

  def _handle_request_noblock(self):
    """Process a single http request."""
    try:
      request, client_address = self.get_request()
    except socket.error:
      return
    if self.verify_request(request, client_address):
      try:
        self.process_request(request, client_address)
      except:
        self.close_request(request)


class RequestHandler(http.server.BaseHTTPRequestHandler):
  """Handler for get requests to test cases."""

  def do_GET(self):  # pylint: disable=invalid-name
    """Handle a GET request."""
    absolute_path = get_absolute_testcase_file(self.path)
    if not absolute_path:
      self.send_response(404)
      self.end_headers()
      return

    try:
      with open(absolute_path) as file_handle:
        data = file_handle.read()
    except IOError:
      self.send_response(403)
      self.end_headers()
      return

    self.send_response(200, 'OK')

    # Send a content type header if applicable.
    mime_type = guess_mime_type(absolute_path)
    if mime_type:
      self.send_header('Content-type', mime_type)

    self.end_headers()
    self.wfile.write(data)

  def log_message(self, fmt, *args):  # pylint: disable=arguments-differ
    """Do not output a log entry to stderr for every request made."""


def run_server(host, port):
  """Run the HTTP server on the given port."""
  httpd = BotHTTPServer((host, port), RequestHandler)
  httpd.serve_forever()


def port_is_open(host, port):
  socket_handle = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
  result = socket_handle.connect_ex((host, port))
  socket_handle.close()
  return result == 0


def start_server_thread(host, port):
  server = threading.Thread(target=run_server, args=(host, port))
  server.daemon = True
  server.start()


def start():
  """Initialize the HTTP server on the specified ports."""
  http_host = 'localhost'
  http_port_1 = environment.get_value('HTTP_PORT_1', 8000)
  http_port_2 = environment.get_value('HTTP_PORT_2', 8080)
  if not port_is_open(http_host, http_port_1):
    start_server_thread(http_host, http_port_1)
  if not port_is_open(http_host, http_port_2):
    start_server_thread(http_host, http_port_2)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for build_setup_host."""

import os
import unittest

from clusterfuzz._internal.bot.untrusted_runner import build_setup_host
from clusterfuzz._internal.protos import untrusted_runner_pb2
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers


def _mock_set_environment_vars(_):
  """Mock build_manager.set_environment_vars."""
  os.environ['APP_PATH'] = ''
  os.environ['APP_DIR'] = ''


class BuildSetupHostTest(unittest.TestCase):
  """Tests for build setup host (trusted side)."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.build_management.build_manager.set_environment_vars',
        'clusterfuzz._internal.bot.untrusted_runner.host.stub',
    ])

    test_helpers.patch_environ(self)

  def test_remote_regular_build(self):
    """Test RemoteRegularBuild."""
    self.mock.stub().SetupRegularBuild.return_value = (
        untrusted_runner_pb2.SetupBuildResponse(
            result=True,
            app_path='/release/bin/app',
            app_path_debug='',
            app_dir='/release/bin',
            build_dir='/release',
            build_url='https://build/url.zip'))

    build = build_setup_host.RemoteRegularBuild('/', 1337,
                                                'https://build/url.zip')
    self.assertTrue(build.setup())
    self.assertEqual(os.environ['APP_PATH'], '/release/bin/app')
    self.assertEqual(os.environ['APP_PATH_DEBUG'], '')
    self.assertEqual(os.environ['APP_DIR'], '/release/bin')
    self.assertEqual(os.environ['BUILD_DIR'], '/release')
    self.assertEqual(os.environ['BUILD_URL'], 'https://build/url.zip')
    self.assertEqual(os.environ['APP_REVISION'], '1337')

    self.mock.stub().SetupRegularBuild.return_value = (
        untrusted_runner_pb2.SetupBuildResponse(result=False))
    self.assertFalse(build.setup())
    self.assertFalse(os.getenv('APP_PATH'))
    self.assertFalse(os.getenv('APP_PATH_DEBUG'))
    self.assertFalse(os.getenv('APP_DIR'))
    self.assertFalse(os.getenv('BUILD_DIR'))
    self.assertFalse(os.getenv('BUILD_URL'))
    self.assertFalse(os.getenv('APP_REVISION'))

  def test_set_env_vars_from_fuzzer(self):
    """Test setting APP_PATH etc from FUZZER_DIR as a fallback."""
    os.environ['FUZZER_DIR'] = '/fuzzer_dir'

    self.mock.stub().SetupRegularBuild.return_value = (
        untrusted_runner_pb2.SetupBuildResponse(
            result=True,
            app_path='',
            app_path_debug='',
            app_dir='',
            build_dir='/release',
            build_url='https://build/url.zip'))
    self.mock.set_environment_vars.side_effect = _mock_set_environment_vars

    build = build_setup_host.RemoteRegularBuild('/', 1337,
                                                'https://build/url.zip')
    self.assertTrue(build.setup())
    self.assertEqual(os.getenv('APP_PATH'), '')
    self.assertEqual(os.getenv('APP_DIR'), '')
    self.assertEqual(os.getenv('BUILD_DIR'), '/release')
    self.assertEqual(os.getenv('BUILD_URL'), 'https://build/url.zip')
    self.assertEqual(os.getenv('APP_REVISION'), '1337')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Running processes with minijail."""

from collections import namedtuple
import os
import shutil
import signal
import subprocess
import tempfile

from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import new_process
from clusterfuzz._internal.system import shell


def _get_minijail_path():
  """Get the minijail path.

  Returns:
    The path to the minijail binary.
  """
  return os.path.join(environment.get_platform_resources_directory(),
                      'minijail0')


def _get_minijail_user_namespace_args():
  """Get user namespace arguments for minijail.

  Returns:
    A list representing arguments to minijail.
  """
  arguments = ['-U']  # User namespace option

  # root (uid 0 in namespace) -> USER.
  # The reason for this is that minijail does setresuid(0, 0, 0) before doing a
  # chroot, which means uid 0 needs access to the chroot dir (owned by USER).
  #
  # Note that we also run fuzzers as uid 0 (but with no capabilities in
  # permitted/effective/inherited sets which *should* mean there's nothing
  # special about it). This is because the uid running the fuzzer also needs
  # access to things owned by USER (fuzzer binaries, supporting files), and USER
  # can only be mapped once.
  uid_map = [
      '0 {0} 1'.format(os.getuid()),
  ]
  arguments.extend(['-m', ','.join(uid_map)])

  return arguments


def _create_chroot_dir(base_dir):
  """Create dir for chroot."""
  return tempfile.mkdtemp(dir=base_dir)


def _create_tmp_mount(base_dir):
  """Create a tmp mount in base_dir."""
  return tempfile.mkdtemp(dir=base_dir)


ChrootBinding = namedtuple('ChrootBinding',
                           ['src_path', 'dest_path', 'writeable'])


class MinijailChroot(object):
  """Minijail environment."""

  # Default directories to bind from host to chroot. Mostly library directories.
  DEFAULT_BINDINGS = [
      '/lib',
      '/lib32',
      '/lib64',
      '/usr/lib',
      '/usr/lib32',
  ]

  def __init__(self, base_dir=None, bindings=None, use_existing_base=False):
    """Inits the MinijailChroot.

    Args:
      base_dir: The directory to create the chroot directory in.
      bindings: Additional bindings (ChrootBinding).
      use_existing_base: use existing base_dir or create a new one.
    """
    if not use_existing_base:
      self._chroot_dir = _create_chroot_dir(base_dir=base_dir)
    else:
      self._chroot_dir = base_dir

    # Create /tmp, /proc directories.
    os.mkdir(os.path.join(self._chroot_dir, 'tmp'))
    os.mkdir(os.path.join(self._chroot_dir, 'proc'))

    # Create a /tmp binding.
    self._tmp_mount = _create_tmp_mount(base_dir=base_dir)
    self._bindings = [
        ChrootBinding(self._tmp_mount, '/tmp', True),
    ]

    self._create_devices()

    # Default dirs. Read-only.
    for directory in self.DEFAULT_BINDINGS:
      if not os.path.exists(directory):
        # Ignore default dirs that don't exist. May happen because of different
        # directory layouts.
        continue

      self.add_binding(ChrootBinding(directory, directory, False))

    if not bindings:
      return

    # Additional bind dirs.
    for binding in bindings:
      self.add_binding(binding)

  def _mknod(self, path, file_type, major, minor):
    """Creates a special file."""
    try:
      with open(os.devnull) as devnull:
        subprocess.check_output(
            [
                'sudo', '-S', 'mknod', '-m', '666', path, file_type,
                str(major),
                str(minor)
            ],
            stdin=devnull,
            stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
      logs.log_error('Failed to call mknod.', output=e.output)

  def _create_devices(self):
    """Create /dev/null, /dev/random, /dev/urandom, and /dev/shm."""
    dev_dir = os.path.join(self._chroot_dir, 'dev')
    os.mkdir(dev_dir)
    self._mknod(os.path.join(dev_dir, 'null'), 'c', 1, 3)
    self._mknod(os.path.join(dev_dir, 'random'), 'c', 1, 8)
    self._mknod(os.path.join(dev_dir, 'urandom'), 'c', 1, 9)
    os.mkdir(os.path.join(dev_dir, 'shm'))

  def _makedirs(self, directory):
    """Create directories for binding in chroot.

    Args:
      directory: The absolute path to the directory target in the chroot.
    """
    if directory[0] == '/':
      directory = directory[1:]

    shell.create_directory(
        os.path.join(self._chroot_dir, directory), create_intermediates=True)

  @property
  def bindings(self):
    return self._bindings

  @property
  def directory(self):
    return self._chroot_dir

  @property
  def tmp_directory(self):
    return self._tmp_mount

  def add_binding(self, binding):
    """Adds a directory to be bound to the chroot.

    Args:
      binding: A ChrootBinding.
    """
    if binding in self._bindings:
      return

    self._makedirs(binding.dest_path)
    self._bindings.append(binding)

  def get_binding(self, src_path):
    """Returns binding for src_path.

    Args:
      src_path: The source directory path.

    Returns:
      A ChrootBinding with the same src_path.
    """
    return next((x for x in self._bindings
                 if os.path.abspath(x.src_path) == os.path.abspath(src_path)),
                None)

  def close(self):
    """Cleanup the chroot environment."""
    shutil.rmtree(self._chroot_dir, ignore_errors=True)
    shutil.rmtree(self._tmp_mount, ignore_errors=True)

  def remove_binding(self, binding):
    """Remove a directory bound to the chroot. This function does not delete the
    given directory.

    Args:
      binding: A ChrootBinding.
    """
    self._bindings.remove(binding)

  def __enter__(self):
    """Context manager override."""
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    """Context manager override."""
    self.close()


class ChromeOSChroot(MinijailChroot):
  """Minijail environment for ChromeOS fuzzers."""

  DEFAULT_BINDINGS = []

  def __init__(self, chroot_dir, bindings=None):
    # Do clean up in case close() was not called.
    self.remove_created_dirs(chroot_dir)
    super(ChromeOSChroot, self).__init__(
        chroot_dir, bindings, use_existing_base=True)

  def remove_created_dirs(self, chroot_dir, minijail_created_dirs=None):
    if minijail_created_dirs is None:
      minijail_created_dirs = ['tmp', 'proc', 'dev']

    for directory in minijail_created_dirs:
      shell.remove_directory(os.path.join(chroot_dir, directory))

  def remove_binding(self, binding):
    """Overriden version of remove_binding that ensures the bound directory is
    removed. This is necessary because unlike in regular minijails, we do not
    delete the entire chroot directory."""
    abs_path = os.path.join(self._chroot_dir, binding.dest_path[1:])
    shell.remove_directory(abs_path)
    super(ChromeOSChroot, self).remove_binding(binding)

  def close(self):
    """Overrides MinijailChroot.close(). Closes the chroot environment. Does
    not delete the chroot directory."""
    for binding in self._bindings:
      self.remove_binding(binding)
    shutil.rmtree(self._tmp_mount, ignore_errors=True)


class MinijailChildProcess(new_process.ChildProcess):
  """Minijail child process."""

  def __init__(self, popen, command, max_stdout_len, stdout_file,
               jailed_pid_file):
    super(MinijailChildProcess, self).__init__(popen, command, max_stdout_len,
                                               stdout_file)
    self._jailed_pid_file = jailed_pid_file

  def terminate(self):
    """Send SIGTERM to the jailed process."""
    self._jailed_pid_file.seek(0)
    jailed_pid = int(self._jailed_pid_file.read())
    os.kill(jailed_pid, signal.SIGTERM)

  def kill(self):
    """Kill minijail and all child processes."""
    os.killpg(self.popen.pid, signal.SIGKILL)


class MinijailProcessRunner(new_process.ProcessRunner):
  """ProcessRunner wrapper for minijail."""

  MINIJAIL_ARGS = [
      '-T',
      'static',  # don't use preload.
      '-c',
      '0',  # drop all capabilities.
      '-n',  # no_new_privs
      '-v',  # mount namespace
      '-p',  # PID namespace
      '-l',  # IPC namespace
      '-I',  # Run jailed process as init.
      '-k',
      'proc,/proc,proc,1'  # Mount procfs RO (1 == MS_RDONLY).
  ]

  PATH_ENVIRONMENT_VALUE = '/bin:/usr/bin'

  def __init__(self, chroot, executable_path, default_args=None):
    super(MinijailProcessRunner, self).__init__(
        executable_path, default_args=default_args)
    self._chroot = chroot

  @property
  def chroot(self):
    return self._chroot

  def get_command(self, additional_args=None):
    """ProcessRunner.get_command override to prepend minijail."""
    base_command = super(MinijailProcessRunner,
                         self).get_command(additional_args)
    command = [_get_minijail_path()]
    command.extend(_get_minijail_user_namespace_args())
    command.extend(self.MINIJAIL_ARGS)

    # Change root filesystem to the chroot directory. See pivot_root(2).
    command.extend(['-P', self._chroot.directory])

    # Bind dirs in chroot.
    for (directory, target, writeable) in self._chroot.bindings:
      command.extend(['-b', '%s,%s,%d' % (directory, target, int(writeable))])

    command.extend(base_command)
    return command

  def run(self,
          additional_args=None,
          max_stdout_len=None,
          extra_env=None,
          stdin=subprocess.PIPE,
          stdout=subprocess.PIPE,
          stderr=subprocess.STDOUT,
          **popen_args):
    """ProcessRunner.run override."""

    pid_file = tempfile.NamedTemporaryFile()
    command = self.get_command(additional_args)

    if stdout == subprocess.PIPE and max_stdout_len:
      stdout = tempfile.TemporaryFile()

    # Insert arguments to write jailed PID to a file.
    command.insert(1, '-f')
    command.insert(2, pid_file.name)

    passed_env = popen_args.pop('env', None)
    from clusterfuzz._internal.bot.untrusted_runner import \
        environment as untrusted_environment
    env = untrusted_environment.get_env_for_untrusted_process(passed_env)
    if extra_env is not None:
      env.update(extra_env)

    env['PATH'] = self.PATH_ENVIRONMENT_VALUE

    return MinijailChildProcess(
        subprocess.Popen(
            command,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            close_fds=True,
            env=env,
            **popen_args),
        command,
        max_stdout_len=max_stdout_len,
        stdout_file=stdout,
        jailed_pid_file=pid_file)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The initialization script for Mac. It is run before running a task."""

import os
import re
import shutil
import subprocess

from clusterfuzz._internal.bot.init_scripts import init_runner

# Example: ('Path: /var/folders/bg/tn9j_qb532s4fz11rzz7m6sc0000gm/0'
#           '//com.apple.LaunchServices-134500.csstore')
LAUNCH_SERVICE_PATH_REGEX = re.compile('^Path: (.+)$')
LSREGISTER_CMD = ('/System/Library/Frameworks/CoreServices.framework'
                  '/Frameworks/LaunchServices.framework/Versions/A/Support'
                  '/lsregister -dump')


def _execute(cmd):
  """Execute command and return output as an iterator."""
  proc = subprocess.Popen(
      cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
  try:
    for line in iter(proc.stdout.readline, b''):
      yield line.decode('utf-8')
  finally:
    proc.kill()


def get_launch_service_path():
  """Get launch service path from lsregister."""
  for line in _execute(LSREGISTER_CMD):
    m = LAUNCH_SERVICE_PATH_REGEX.match(line)
    if not m:
      continue

    return '/'.join(m.group(1).split('/')[:5])

  return None


def clear_launch_service_data():
  """See crbug.com/661221 for more info."""
  path = get_launch_service_path()
  if not path or not os.path.exists(path):
    return
  # Best effort removal. We use shutil instead of shell.remove_directory since
  # it's too noisy and there are many files that cannot be removed.
  shutil.rmtree(os.path.join(path, '0'), ignore_errors=True)
  shutil.rmtree(os.path.join(path, 'T'), ignore_errors=True)


def run():
  """Run the initialization for Mac."""
  init_runner.run()
  clear_launch_service_data()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""mark_security tests."""
import unittest

import flask
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.testcase_detail import mark_security
from libs import form


@test_utils.with_cloud_emulators('datastore')
class HandlerTest(unittest.TestCase):
  """Test Handler."""

  def setUp(self):
    test_helpers.patch(self, [
        'libs.auth.get_current_user',
        'libs.auth.is_current_user_admin',
        'handlers.testcase_detail.show.get_testcase_detail',
    ])
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=mark_security.Handler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    self.testcase = data_types.Testcase()
    self.testcase.put()
    self.mock.is_current_user_admin.return_value = True
    self.mock.get_testcase_detail.return_value = {'testcase': 'yes'}
    self.mock.get_current_user().email = 'test@user.com'

  def test_succeed(self):
    """Mark a testcase as security-related."""
    self.testcase.security_flag = False
    self.testcase.put()

    resp = self.app.post_json(
        '/', {
            'testcaseId': self.testcase.key.id(),
            'csrf_token': form.generate_csrf_token(),
            'security': True,
            'severity': data_types.SecuritySeverity.HIGH,
        })

    self.assertEqual(200, resp.status_int)
    self.assertEqual('yes', resp.json['testcase'])

    testcase = self.testcase.key.get()
    self.assertTrue(testcase.security_flag)
    self.assertEqual(data_types.SecuritySeverity.HIGH,
                     testcase.security_severity)

  def test_succeed_non_security(self):
    """Mark a testcase as not security related."""
    self.testcase.security_flag = True
    self.testcase.put()

    resp = self.app.post_json(
        '/', {
            'testcaseId': self.testcase.key.id(),
            'csrf_token': form.generate_csrf_token(),
            'security': False,
        })

    self.assertEqual(200, resp.status_int)
    self.assertEqual('yes', resp.json['testcase'])

    testcase = self.testcase.key.get()
    self.assertFalse(testcase.security_flag)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fuzzers.utils."""

import os
import shutil
import tempfile
import unittest

from clusterfuzz._internal.bot.fuzzers import utils
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers


class IsFuzzTargetLocalTest(unittest.TestCase):
  """is_fuzz_target_local tests."""

  def setUp(self):
    test_helpers.patch_environ(self)
    self.temp_dir = tempfile.mkdtemp()

  def tearDown(self):
    shutil.rmtree(self.temp_dir, ignore_errors=True)

  def _create_file(self, name, contents=b''):
    path = os.path.join(self.temp_dir, name)
    with open(path, 'wb') as f:
      f.write(contents)

    return path

  def test_not_a_fuzzer_invalid_name(self):
    path = self._create_file('abc$_fuzzer', contents=b'LLVMFuzzerTestOneInput')
    self.assertFalse(utils.is_fuzz_target_local(path))

  def test_not_a_fuzzer_blocklisted_name(self):
    path = self._create_file(
        'jazzer_driver', contents=b'LLVMFuzzerTestOneInput')
    self.assertFalse(utils.is_fuzz_target_local(path))

  def test_not_a_fuzzer_without_extension(self):
    path = self._create_file('abc', contents=b'anything')
    self.assertFalse(utils.is_fuzz_target_local(path))

  def test_not_a_fuzzer_with_extension(self):
    path = self._create_file('abc.dict', contents=b'LLVMFuzzerTestOneInput')
    self.assertFalse(utils.is_fuzz_target_local(path))

  def test_not_a_fuzzer_with_extension_and_suffix(self):
    path = self._create_file(
        'abc_fuzzer.dict', contents=b'LLVMFuzzerTestOneInput')
    self.assertFalse(utils.is_fuzz_target_local(path))

  def test_fuzzer_posix(self):
    path = self._create_file('abc_fuzzer', contents=b'anything')
    self.assertTrue(utils.is_fuzz_target_local(path))

  def test_fuzzer_win(self):
    path = self._create_file('abc_fuzzer.exe', contents=b'anything')
    self.assertTrue(utils.is_fuzz_target_local(path))

  def test_fuzzer_py(self):
    path = self._create_file('abc_fuzzer.par', contents=b'anything')
    self.assertTrue(utils.is_fuzz_target_local(path))

  def test_fuzzer_not_exist(self):
    self.assertFalse(utils.is_fuzz_target_local('/not_exist_fuzzer'))

  def test_fuzzer_without_suffix(self):
    path = self._create_file(
        'abc', contents=b'anything\nLLVMFuzzerTestOneInput')
    self.assertTrue(utils.is_fuzz_target_local(path))

  def test_fuzzer_with_name_regex_match(self):
    environment.set_value('FUZZER_NAME_REGEX', '.*_custom$')
    path = self._create_file('a_custom', contents=b'anything')
    self.assertTrue(utils.is_fuzz_target_local(path))

  def test_fuzzer_with_file_string_and_without_name_regex_match(self):
    environment.set_value('FUZZER_NAME_REGEX', '.*_custom$')
    path = self._create_file(
        'nomatch', contents=b'anything\nLLVMFuzzerTestOneInput')
    self.assertFalse(utils.is_fuzz_target_local(path))

  def test_fuzzer_without_file_string_and_without_name_regex_match(self):
    environment.set_value('FUZZER_NAME_REGEX', '.*_custom$')
    path = self._create_file('nomatch', contents=b'anything')
    self.assertFalse(utils.is_fuzz_target_local(path))

  def test_fuzzer_with_fuzzer_name_and_without_name_regex_match(self):
    environment.set_value('FUZZER_NAME_REGEX', '.*_custom$')
    path = self._create_file(
        'a_fuzzer', contents=b'anything\nLLVMFuzzerTestOneInput')
    self.assertTrue(utils.is_fuzz_target_local(path))

  def test_file_handle(self):
    """Test with a file handle."""
    path = self._create_file(
        'abc', contents=b'anything\nLLVMFuzzerTestOneInput')
    with open(path, 'rb') as f:
      self.assertTrue(utils.is_fuzz_target_local('name', f))


class GetSupportingFileTest(unittest.TestCase):
  """Tests for get_supporting_file."""

  def test_no_extension(self):
    """Test no extension."""
    self.assertEqual('/a/b.labels', utils.get_supporting_file(
        '/a/b', '.labels'))

  def test_unknown_extension(self):
    """Test unknown extension."""
    self.assertEqual('/a/b.c.labels',
                     utils.get_supporting_file('/a/b.c', '.labels'))

  def test_exe(self):
    """Test exe extension."""
    self.assertEqual('/a/b.labels',
                     utils.get_supporting_file('/a/b.exe', '.labels'))

  def test_par(self):
    """Test par extension."""
    self.assertEqual('/a/b.labels',
                     utils.get_supporting_file('/a/b.par', '.labels'))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helpers for sending mail."""
from sendgrid import SendGridAPIClient
from sendgrid.helpers.mail import From
from sendgrid.helpers.mail import HtmlContent
from sendgrid.helpers.mail import Mail
from sendgrid.helpers.mail import Subject
from sendgrid.helpers.mail import To

from clusterfuzz._internal.config import db_config
from clusterfuzz._internal.metrics import logs


def send(to_email, subject, html_content):
  """Send email."""
  sendgrid_api_key = db_config.get_value('sendgrid_api_key')
  if not sendgrid_api_key:
    logs.log_warn('Skipping email as SendGrid API key is not set in config.')
    return

  from_email = db_config.get_value('sendgrid_sender')
  if not from_email:
    logs.log_warn('Skipping email as SendGrid sender is not set in config.')
    return

  message = Mail(
      from_email=From(str(from_email)),
      to_emails=To(str(to_email)),
      subject=Subject(subject),
      html_content=HtmlContent(str(html_content)))
  try:
    sg = SendGridAPIClient(sendgrid_api_key)
    response = sg.send(message)
    logs.log(
        'Sent email to %s.' % to_email,
        status_code=response.status_code,
        body=response.body,
        headers=response.headers)
  except Exception:
    logs.log_error('Failed to send email to %s.' % to_email)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Constants that are meaningful to libFuzzer.
Should not have any dependencies.
Note that libFuzzers arguments take the form -flag=value. Thus any variables
defined in this function that end with the suffix "_FLAG" should contain
"-flag=". Any variable that ends with the suffix "_ARGUMENT" should contain
"-flag=value".
"""

# libFuzzer flags.
ARTIFACT_PREFIX_FLAG = '-artifact_prefix='

DATA_FLOW_TRACE_FLAG = '-data_flow_trace='

DICT_FLAG = '-dict='

FOCUS_FUNCTION_FLAG = '-focus_function='

FORK_FLAG = '-fork='

MAX_LEN_FLAG = '-max_len='

MAX_TOTAL_TIME_FLAG = '-max_total_time='

RSS_LIMIT_FLAG = '-rss_limit_mb='

RUNS_FLAG = '-runs='

TIMEOUT_FLAG = '-timeout='

EXACT_ARTIFACT_PATH_FLAG = '-exact_artifact_path='

# libFuzzer arguments.
ANALYZE_DICT_ARGUMENT = '-analyze_dict=1'

CLEANSE_CRASH_ARGUMENT = '-cleanse_crash=1'

MERGE_ARGUMENT = '-merge=1'

MERGE_CONTROL_FILE_ARGUMENT = '-merge_control_file='

MINIMIZE_CRASH_ARGUMENT = '-minimize_crash=1'

PRINT_FINAL_STATS_ARGUMENT = '-print_final_stats=1'

TMP_ARTIFACT_PREFIX_ARGUMENT = '/tmp/'

VALUE_PROFILE_ARGUMENT = '-use_value_profile=1'

# Default value for rss_limit_mb flag to catch OOM.s
DEFAULT_RSS_LIMIT_MB = 2560

# Default value for timeout flag to catch timeouts.
DEFAULT_TIMEOUT_LIMIT = 25

# Buffer for processing crash reports.
REPORT_PROCESSING_TIME = 5

# libFuzzer's exit code if a bug occurred in libFuzzer.
LIBFUZZER_ERROR_EXITCODE = 1

# Defines value of runs argument when loading a testcase.
RUNS_TO_REPRODUCE = 100

# libFuzzer's exit code if a bug was found in the target code.
TARGET_ERROR_EXITCODE = 77



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler used for setting up oss-fuzz jobs."""

import base64
import collections
import copy
import json
import re

from google.cloud import ndb
import requests
import six
import yaml

from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.base import untrusted
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.config import db_config
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_utils
from clusterfuzz._internal.fuzzing import fuzzer_selection
from clusterfuzz._internal.google_cloud_utils import pubsub
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from handlers import base_handler
from libs import handler

from . import service_accounts

BUILD_BUCKET_PATH_TEMPLATE = (
    'gs://%BUCKET%/%PROJECT%/%PROJECT%-%SANITIZER%-([0-9]+).zip')

BACKUPS_LIFECYCLE = storage.generate_life_cycle_config('Delete', age=100)
LOGS_LIFECYCLE = storage.generate_life_cycle_config('Delete', age=14)
QUARANTINE_LIFECYCLE = storage.generate_life_cycle_config('Delete', age=90)

JOB_TEMPLATE = ('{build_type} = {build_bucket_path}\n'
                'PROJECT_NAME = {project}\n'
                'SUMMARY_PREFIX = {project}\n'
                'MANAGED = True\n')

OBJECT_VIEWER_IAM_ROLE = 'roles/storage.objectViewer'
OBJECT_ADMIN_IAM_ROLE = 'roles/storage.objectAdmin'

VALID_PROJECT_NAME_REGEX = re.compile(r'^[a-zA-Z0-9_-]+$')

REVISION_URL = ('https://commondatastorage.googleapis.com/'
                '{bucket}/{project}/{project}-{sanitizer}-%s.srcmap.json')

REQUEST_TIMEOUT = 60

ALLOWED_VIEW_RESTRICTIONS = ['none', 'security', 'all']

PUBSUB_PLATFORMS = ['linux']

MEMORY_SAFE_LANGUAGES = {'go', 'java', 'python', 'rust'}
OSS_FUZZ_DEFAULT_PROJECT_CPU_WEIGHT = 1.0
OSS_FUZZ_MEMORY_SAFE_LANGUAGE_PROJECT_WEIGHT = 0.2

SetupResult = collections.namedtuple('SetupResult', 'project_names job_names')


class ProjectSetupError(Exception):
  """Exception."""


class JobInfo(object):
  """Job information."""

  def __init__(self,
               prefix,
               engine,
               memory_tool,
               cf_job_templates,
               architecture='x86_64',
               experimental=False,
               minimize_job_override=None):
    self.prefix = prefix
    self.engine = engine
    self.memory_tool = memory_tool
    self.architecture = architecture
    self.cf_job_templates = cf_job_templates
    self.experimental = experimental
    self.minimize_job_override = minimize_job_override

  def job_name(self, project_name, config_suffix):
    return (
        self.prefix + data_types.normalized_name(project_name) + config_suffix)


# The order of templates is important here. Later templates override settings in
# the earlier ones. An engine template may override vars set for a sanitizer.
LIBFUZZER_ASAN_JOB = JobInfo('libfuzzer_asan_', 'libfuzzer', 'address',
                             ['libfuzzer', 'engine_asan', 'prune'])
LIBFUZZER_MSAN_JOB = JobInfo('libfuzzer_msan_', 'libfuzzer', 'memory',
                             ['libfuzzer', 'engine_msan'])
LIBFUZZER_UBSAN_JOB = JobInfo('libfuzzer_ubsan_', 'libfuzzer', 'undefined',
                              ['libfuzzer', 'engine_ubsan'])
LIBFUZZER_ASAN_I386_JOB = JobInfo(
    'libfuzzer_asan_i386_',
    'libfuzzer',
    'address', ['libfuzzer', 'engine_asan'],
    architecture='i386')

AFL_ASAN_JOB = JobInfo(
    'afl_asan_',
    'afl',
    'address', ['afl', 'engine_asan'],
    minimize_job_override=LIBFUZZER_ASAN_JOB)
NO_ENGINE_ASAN_JOB = JobInfo('asan_', 'none', 'address', [])

HONGGFUZZ_ASAN_JOB = JobInfo(
    'honggfuzz_asan_',
    'honggfuzz',
    'address', ['honggfuzz', 'engine_asan'],
    minimize_job_override=LIBFUZZER_ASAN_JOB)

GFT_ASAN_JOB = JobInfo('googlefuzztest_asan_', 'googlefuzztest', 'address',
                       ['googlefuzztest', 'engine_asan'])
GFT_MSAN_JOB = JobInfo('googlefuzztest_msan_', 'googlefuzztest', 'memory',
                       ['googlefuzztest', 'engine_msan'])
GFT_UBSAN_JOB = JobInfo('googlefuzztest_ubsan_', 'googlefuzztest', 'undefined',
                        ['googlefuzztest', 'engine_ubsan'])

JOB_MAP = {
    'libfuzzer': {
        'x86_64': {
            'address': LIBFUZZER_ASAN_JOB,
            'memory': LIBFUZZER_MSAN_JOB,
            'undefined': LIBFUZZER_UBSAN_JOB,
        },
        'i386': {
            'address': LIBFUZZER_ASAN_I386_JOB,
        },
    },
    'afl': {
        'x86_64': {
            'address': AFL_ASAN_JOB,
        }
    },
    'honggfuzz': {
        'x86_64': {
            'address': HONGGFUZZ_ASAN_JOB,
        },
    },
    'googlefuzztest': {
        'x86_64': {
            'address': GFT_ASAN_JOB,
            'memory': GFT_MSAN_JOB,
            'undefined': GFT_UBSAN_JOB,
        },
    },
    'none': {
        'x86_64': {
            'address': NO_ENGINE_ASAN_JOB,
        }
    }
}

DEFAULT_ARCHITECTURES = ['x86_64']
DEFAULT_SANITIZERS = ['address', 'undefined']
DEFAULT_ENGINES = ['libfuzzer', 'afl', 'honggfuzz']


def _to_experimental_job(job_info):
  job_info = copy.copy(job_info)
  job_info.experimental = True
  return job_info


def get_github_url(url):
  """Return contents of URL."""
  github_credentials = db_config.get_value('github_credentials')
  if not github_credentials:
    raise ProjectSetupError('No github credentials.')

  client_id, client_secret = github_credentials.strip().split(';')
  response = requests.get(url, auth=(client_id, client_secret))
  if response.status_code != 200:
    logs.log_error(
        'Failed to get github url: %s' % url, status_code=response.status_code)
    response.raise_for_status()

  return json.loads(response.text)


def find_github_item_url(github_json, name):
  """Get url of a blob/tree from a github json response."""
  for item in github_json['tree']:
    if item['path'] == name:
      return item['url']

  return None


def get_oss_fuzz_projects():
  """Return list of projects for oss-fuzz."""
  ossfuzz_tree_url = ('https://api.github.com/repos/google/oss-fuzz/'
                      'git/trees/master')
  tree = get_github_url(ossfuzz_tree_url)
  projects = []

  projects_url = find_github_item_url(tree, 'projects')
  if not projects_url:
    logs.log_error('No projects found.')
    return []

  tree = get_github_url(projects_url)
  for item in tree['tree']:
    if item['type'] != 'tree':
      continue

    item_json = get_github_url(item['url'])
    project_yaml_url = find_github_item_url(item_json, 'project.yaml')
    if not project_yaml_url:
      continue

    projects_yaml = get_github_url(project_yaml_url)
    info = yaml.safe_load(base64.b64decode(projects_yaml['content']))

    has_dockerfile = (
        find_github_item_url(item_json, 'Dockerfile') or 'dockerfile' in info)
    if not has_dockerfile:
      continue

    projects.append((item['path'], info))

  return projects


def get_projects_from_gcs(gcs_url):
  """Get projects from GCS path."""
  data = json.loads(storage.read_data(gcs_url))
  return [(project['name'], project) for project in data['projects']]


def _process_sanitizers_field(sanitizers):
  """Pre-process sanitizers field into a map from sanitizer name -> dict of
  options."""
  processed_sanitizers = {}
  if not isinstance(sanitizers, list):
    return None

  # each field can either be a Map or a String:
  # sanitizers:
  #   - undefined:
  #       experimental: true
  #   - address
  #   - memory
  for sanitizer in sanitizers:
    if isinstance(sanitizer, str):
      processed_sanitizers[sanitizer] = {}
    elif isinstance(sanitizer, dict):
      for key, value in six.iteritems(sanitizer):
        processed_sanitizers[key] = value
    else:
      return None

  return processed_sanitizers


def get_jobs_for_project(project, info):
  """Return jobs for the project."""
  sanitizers = _process_sanitizers_field(
      info.get('sanitizers', DEFAULT_SANITIZERS))
  if not sanitizers:
    logs.log_error('Invalid sanitizers field for %s.' % project)
    return []

  engines = info.get('fuzzing_engines', DEFAULT_ENGINES)
  architectures = info.get('architectures', DEFAULT_ARCHITECTURES)

  jobs = []
  for engine in engines:
    if engine not in JOB_MAP:
      continue

    for architecture in architectures:
      if architecture not in JOB_MAP[engine]:
        continue

      for sanitizer, options in six.iteritems(sanitizers):
        experimental = (
            options.get('experimental', False) or
            info.get('experimental', False))
        if sanitizer in JOB_MAP[engine][architecture]:
          job = JOB_MAP[engine][architecture][sanitizer]
          if experimental:
            job = _to_experimental_job(job)

          jobs.append(job)

  return jobs


def convert_googlemail_to_gmail(email):
  """Convert @googlemail.com to @gmail.com."""
  # TODO(ochang): Investiate if we can/need to do this in general, and not just
  # for cloud storage bucket IAMs.
  if email.endswith('@googlemail.com'):
    return email.split('@')[0] + '@gmail.com'

  return email


def _add_users_to_bucket(info, client, bucket_name, iam_policy):
  """Add user account to bucket."""
  ccs = sorted(
      ['user:' + convert_googlemail_to_gmail(cc) for cc in ccs_from_info(info)])
  binding = storage.get_bucket_iam_binding(iam_policy, OBJECT_VIEWER_IAM_ROLE)

  if binding:
    # buckets.getIamPolicy can return duplicate members when we add a @gmail.com
    # as well as @googlemail.com address for the same account.
    binding['members'] = sorted(list(set(binding['members'])))
    if binding['members'] == ccs:
      return iam_policy

    filtered_members = [
        member for member in binding['members'] if member in ccs
    ]

    if len(filtered_members) != len(binding['members']):
      # Remove old members.
      binding['members'] = filtered_members
      iam_policy = storage.set_bucket_iam_policy(client, bucket_name,
                                                 iam_policy)

  # We might have no binding either from start or after filtering members above.
  # Create a new one in those cases.
  binding = storage.get_or_create_bucket_iam_binding(iam_policy,
                                                     OBJECT_VIEWER_IAM_ROLE)

  for cc in ccs:
    if cc in binding['members']:
      continue

    logs.log('Adding %s to bucket IAM for %s' % (cc, bucket_name))
    # Add CCs one at a time since the API does not work with invalid or
    # non-Google emails.
    modified_iam_policy = storage.add_single_bucket_iam(
        client, iam_policy, OBJECT_VIEWER_IAM_ROLE, bucket_name, cc)
    if modified_iam_policy:
      iam_policy = modified_iam_policy
      binding = storage.get_bucket_iam_binding(iam_policy,
                                               OBJECT_VIEWER_IAM_ROLE)

  if not binding['members']:
    # Check that the final binding has members. Empty bindings are not valid.
    storage.remove_bucket_iam_binding(iam_policy, OBJECT_VIEWER_IAM_ROLE)

  return iam_policy


def _set_bucket_service_account(service_account, client, bucket_name,
                                iam_policy):
  """Set service account for a bucket."""
  # Add service account as objectAdmin.
  binding = storage.get_or_create_bucket_iam_binding(iam_policy,
                                                     OBJECT_ADMIN_IAM_ROLE)

  members = ['serviceAccount:' + service_account['email']]
  if members == binding['members']:
    # No changes required.
    return iam_policy

  binding['members'] = members
  return storage.set_bucket_iam_policy(client, bucket_name, iam_policy)


def add_bucket_iams(info, client, bucket_name, service_account):
  """Add CC'ed users to storage bucket IAM."""
  iam_policy = storage.get_bucket_iam_policy(client, bucket_name)
  if not iam_policy:
    return

  iam_policy = _add_users_to_bucket(info, client, bucket_name, iam_policy)
  _set_bucket_service_account(service_account, client, bucket_name, iam_policy)


def add_service_account_to_bucket(client, bucket_name, service_account, role):
  """Add service account to the gcr.io images bucket."""
  iam_policy = storage.get_bucket_iam_policy(client, bucket_name)
  if not iam_policy:
    return

  binding = storage.get_or_create_bucket_iam_binding(iam_policy, role)

  member = 'serviceAccount:' + service_account['email']
  if member in binding['members']:
    # No changes required.
    return

  binding['members'].append(member)
  storage.set_bucket_iam_policy(client, bucket_name, iam_policy)


def ccs_from_info(info):
  """Get list of CC's from project info."""

  def _get_ccs(field_name, allow_list=True):
    """Return list of emails to cc given a field name."""
    if field_name not in info:
      return []

    field_value = info.get(field_name)
    if allow_list and isinstance(field_value, list):
      return field_value
    if isinstance(field_value, str):
      return [field_value]

    raise ProjectSetupError(
        'Bad value for field {field_name}: {field_value}.'.format(
            field_name=field_name, field_value=field_value))

  ccs = []
  ccs.extend(_get_ccs('primary_contact', allow_list=False))
  ccs.extend(_get_ccs('auto_ccs'))
  ccs.extend(_get_ccs('vendor_ccs'))

  return [utils.normalize_email(cc) for cc in ccs]


def update_fuzzer_jobs(fuzzer_entities, job_names):
  """Update fuzzer job mappings."""
  to_delete = []

  for job in data_types.Job.query():
    if not job.environment_string:
      continue

    job_environment = job.get_environment()
    if not utils.string_is_true(job_environment.get('MANAGED', 'False')):
      continue

    if job.name in job_names:
      continue

    logs.log('Deleting job %s' % job.name)
    to_delete.append(job.key)
    for fuzzer_entity in fuzzer_entities:
      try:
        fuzzer_entity.jobs.remove(job.name)
      except ValueError:
        pass

  for fuzzer_entity in fuzzer_entities:
    fuzzer_entity.put()
    fuzzer_selection.update_mappings_for_fuzzer(fuzzer_entity)

  if to_delete:
    ndb_utils.delete_multi(to_delete)


def cleanup_old_projects_settings(project_names):
  """Delete old projects that are no longer used or disabled."""
  to_delete = []

  for project in data_types.OssFuzzProject.query():
    if project.name not in project_names:
      logs.log('Deleting project %s' % project.name)
      to_delete.append(project.key)

  if to_delete:
    ndb_utils.delete_multi(to_delete)


def create_project_settings(project, info, service_account):
  """Setup settings for ClusterFuzz (such as CPU distribution)."""
  key = ndb.Key(data_types.OssFuzzProject, project)
  oss_fuzz_project = key.get()

  # Expecting to run a blackbox fuzzer, so use high end hosts.
  is_high_end = info.get('blackbox', False)

  ccs = ccs_from_info(info)
  language = info.get('language')

  if oss_fuzz_project:
    if oss_fuzz_project.service_account != service_account['email']:
      oss_fuzz_project.service_account = service_account['email']
      oss_fuzz_project.put()

    if oss_fuzz_project.high_end != is_high_end:
      oss_fuzz_project.high_end = is_high_end
      oss_fuzz_project.put()

    if oss_fuzz_project.ccs != ccs:
      oss_fuzz_project.ccs = ccs
      oss_fuzz_project.put()
  else:
    if language in MEMORY_SAFE_LANGUAGES:
      cpu_weight = OSS_FUZZ_MEMORY_SAFE_LANGUAGE_PROJECT_WEIGHT
    else:
      cpu_weight = OSS_FUZZ_DEFAULT_PROJECT_CPU_WEIGHT

    data_types.OssFuzzProject(
        id=project,
        name=project,
        high_end=is_high_end,
        cpu_weight=cpu_weight,
        service_account=service_account['email'],
        ccs=ccs).put()


def create_pubsub_topics(project):
  """Create pubsub topics for tasks."""
  for platform in PUBSUB_PLATFORMS:
    name = untrusted.queue_name(project, platform)
    client = pubsub.PubSubClient()
    application_id = utils.get_application_id()

    topic_name = pubsub.topic_name(application_id, name)
    if client.get_topic(topic_name) is None:
      client.create_topic(topic_name)

    subscription_name = pubsub.subscription_name(application_id, name)
    if client.get_subscription(subscription_name) is None:
      client.create_subscription(subscription_name, topic_name)


def cleanup_pubsub_topics(project_names):
  """Delete old pubsub topics and subscriptions."""
  client = pubsub.PubSubClient()
  application_id = utils.get_application_id()

  expected_topics = set()
  for platform in PUBSUB_PLATFORMS:
    expected_topics.update(
        [untrusted.queue_name(project, platform) for project in project_names])

  pubsub_config = local_config.Config('pubsub.queues')
  unmanaged_queues = [queue['name'] for queue in pubsub_config.get('resources')]

  for topic in client.list_topics(pubsub.project_name(application_id)):
    _, name = pubsub.parse_name(topic)

    if (not name.startswith(tasks.JOBS_PREFIX) and
        not name.startswith(tasks.HIGH_END_JOBS_PREFIX)):
      # Some topic created by another service, ignore.
      continue

    if name in unmanaged_queues:
      continue

    if name in expected_topics:
      continue

    for subscription in client.list_topic_subscriptions(topic):
      client.delete_subscription(subscription)

    client.delete_topic(topic)


class ProjectSetup(object):
  """Project setup."""

  def __init__(self,
               build_bucket_path_template,
               revision_url_template,
               build_type,
               config_suffix='',
               external_config=None,
               segregate_projects=False,
               experimental_sanitizers=None,
               engine_build_buckets=None,
               fuzzer_entities=None,
               add_info_labels=False,
               add_revision_mappings=False,
               additional_vars=None):
    self._build_type = build_type
    self._config_suffix = config_suffix
    self._external_config = external_config
    self._build_bucket_path_template = build_bucket_path_template
    self._revision_url_template = revision_url_template
    self._segregate_projects = segregate_projects
    self._experimental_sanitizers = experimental_sanitizers
    self._engine_build_buckets = engine_build_buckets
    self._fuzzer_entities = fuzzer_entities
    self._add_info_labels = add_info_labels
    self._add_revision_mappings = add_revision_mappings
    self._additional_vars = additional_vars

  def _get_build_bucket(self, engine, architecture):
    """Return the bucket for the given |engine| and |architecture|."""
    if architecture != 'x86_64':
      engine += '-' + architecture

    bucket = self._engine_build_buckets.get(engine)
    if not bucket:
      raise ProjectSetupError('Invalid fuzzing engine ' + engine)

    return bucket

  def _deployment_bucket_name(self):
    """Deployment bucket name."""
    return '{project}-deployment'.format(project=utils.get_application_id())

  def _shared_corpus_bucket_name(self):
    """Shared corpus bucket name."""
    return environment.get_value('SHARED_CORPUS_BUCKET')

  def _mutator_plugins_bucket_name(self):
    """Mutator plugins bucket name."""
    return environment.get_value('MUTATOR_PLUGINS_BUCKET')

  def _backup_bucket_name(self, project_name):
    """Return the backup_bucket_name."""
    return project_name + '-backup.' + data_handler.bucket_domain_suffix()

  def _corpus_bucket_name(self, project_name):
    """Return the corpus_bucket_name."""
    return project_name + '-corpus.' + data_handler.bucket_domain_suffix()

  def _quarantine_bucket_name(self, project_name):
    """Return the quarantine_bucket_name."""
    return project_name + '-quarantine.' + data_handler.bucket_domain_suffix()

  def _logs_bucket_name(self, project_name):
    """Return the logs bucket name."""
    return project_name + '-logs.' + data_handler.bucket_domain_suffix()

  def _create_service_accounts_and_buckets(self, project, info):
    """Create per-project service account and buckets."""
    service_account = service_accounts.get_or_create_service_account(project)
    service_accounts.set_service_account_roles(service_account)

    # Create GCS buckets.
    backup_bucket_name = self._backup_bucket_name(project)
    corpus_bucket_name = self._corpus_bucket_name(project)
    logs_bucket_name = self._logs_bucket_name(project)
    quarantine_bucket_name = self._quarantine_bucket_name(project)

    storage.create_bucket_if_needed(backup_bucket_name, BACKUPS_LIFECYCLE)
    storage.create_bucket_if_needed(corpus_bucket_name)
    storage.create_bucket_if_needed(quarantine_bucket_name,
                                    QUARANTINE_LIFECYCLE)
    storage.create_bucket_if_needed(logs_bucket_name, LOGS_LIFECYCLE)

    client = storage.create_discovery_storage_client()
    try:
      add_bucket_iams(info, client, backup_bucket_name, service_account)
      add_bucket_iams(info, client, corpus_bucket_name, service_account)
      add_bucket_iams(info, client, logs_bucket_name, service_account)
      add_bucket_iams(info, client, quarantine_bucket_name, service_account)
    except Exception as e:
      logs.log_error('Failed to add bucket IAMs for %s: %s' % (project, e))

    # Grant the service account read access to deployment, shared corpus and
    # mutator plugin buckets.
    add_service_account_to_bucket(client, self._deployment_bucket_name(),
                                  service_account, OBJECT_VIEWER_IAM_ROLE)
    add_service_account_to_bucket(client, self._shared_corpus_bucket_name(),
                                  service_account, OBJECT_VIEWER_IAM_ROLE)
    add_service_account_to_bucket(client, self._mutator_plugins_bucket_name(),
                                  service_account, OBJECT_VIEWER_IAM_ROLE)

    data_bundles = {
        fuzzer_entity.data_bundle_name
        for fuzzer_entity in six.itervalues(self._fuzzer_entities)
        if fuzzer_entity.data_bundle_name
    }
    for data_bundle in data_bundles:
      # Workers also need to be able to set up these global bundles.
      data_bundle_bucket_name = data_handler.get_data_bundle_bucket_name(
          data_bundle)
      add_service_account_to_bucket(client, data_bundle_bucket_name,
                                    service_account, OBJECT_VIEWER_IAM_ROLE)

    return (service_account, backup_bucket_name, corpus_bucket_name,
            logs_bucket_name, quarantine_bucket_name)

  def _get_build_bucket_path(self, project_name, info, engine, memory_tool,
                             architecture):
    """Returns the build bucket path for the |project|, |engine|, |memory_tool|,
    and |architecture|."""
    build_path = info.get('build_path')
    if not build_path:
      build_path = self._build_bucket_path_template

    build_path = build_path.replace(
        '%BUCKET%', self._get_build_bucket(engine, architecture))
    build_path = build_path.replace('%PROJECT%', project_name)
    build_path = build_path.replace('%ENGINE%', engine)
    build_path = build_path.replace('%SANITIZER%', memory_tool)
    return build_path

  def _sync_job(self, project, info, corpus_bucket_name, quarantine_bucket_name,
                logs_bucket_name, backup_bucket_name):
    """Sync the config with ClusterFuzz."""
    # Create/update ClusterFuzz jobs.
    job_names = []

    for template in get_jobs_for_project(project, info):
      if template.engine == 'none':
        # Engine-less jobs are not automatically managed.
        continue

      fuzzer_entity = self._fuzzer_entities.get(template.engine)
      if not fuzzer_entity:
        raise ProjectSetupError('Invalid fuzzing engine ' + template.engine)

      job_name = template.job_name(project, self._config_suffix)
      job = data_types.Job.query(data_types.Job.name == job_name).get()
      if not job:
        job = data_types.Job()

      if self._external_config:
        if ('reproduction_topic' not in self._external_config or
            'updates_subscription' not in self._external_config):
          raise ProjectSetupError('Invalid external_config.')

        job.external_reproduction_topic = self._external_config[
            'reproduction_topic']
        job.external_updates_subscription = self._external_config[
            'updates_subscription']
      else:
        job.external_reproduction_topic = None
        job.external_updates_subscription = None

      if not info.get('disabled', False):
        job_names.append(job_name)
        if job_name not in fuzzer_entity.jobs and not job.is_external():
          # Enable new job.
          fuzzer_entity.jobs.append(job_name)

      job.name = job_name
      if self._segregate_projects:
        job.platform = untrusted.platform_name(project, 'linux')
      else:
        # TODO(ochang): Support other platforms?
        job.platform = 'LINUX'

      job.templates = template.cf_job_templates

      job.environment_string = JOB_TEMPLATE.format(
          build_type=self._build_type,
          build_bucket_path=self._get_build_bucket_path(
              project, info, template.engine, template.memory_tool,
              template.architecture),
          engine=template.engine,
          project=project)

      if self._add_revision_mappings:
        revision_vars_url = self._revision_url_template.format(
            project=project,
            bucket=self._get_build_bucket(template.engine,
                                          template.architecture),
            sanitizer=template.memory_tool)

        job.environment_string += (
            'REVISION_VARS_URL = {revision_vars_url}\n'.format(
                revision_vars_url=revision_vars_url))

      if logs_bucket_name:
        job.environment_string += 'FUZZ_LOGS_BUCKET = {logs_bucket}\n'.format(
            logs_bucket=logs_bucket_name)

      if corpus_bucket_name:
        job.environment_string += 'CORPUS_BUCKET = {corpus_bucket}\n'.format(
            corpus_bucket=corpus_bucket_name)

      if quarantine_bucket_name:
        job.environment_string += (
            'QUARANTINE_BUCKET = {quarantine_bucket}\n'.format(
                quarantine_bucket=quarantine_bucket_name))

      if backup_bucket_name:
        job.environment_string += 'BACKUP_BUCKET = {backup_bucket}\n'.format(
            backup_bucket=backup_bucket_name)

      if self._add_info_labels:
        job.environment_string += (
            'AUTOMATIC_LABELS = Proj-{project},Engine-{engine}\n'.format(
                project=project,
                engine=template.engine,
            ))

      help_url = info.get('help_url')
      if help_url:
        job.environment_string += 'HELP_URL = %s\n' % help_url

      if (template.experimental or
          (self._experimental_sanitizers and
           template.memory_tool in self._experimental_sanitizers)):
        job.environment_string += 'EXPERIMENTAL = True\n'

      if template.minimize_job_override:
        minimize_job_override = template.minimize_job_override.job_name(
            project, self._config_suffix)
        job.environment_string += (
            'MINIMIZE_JOB_OVERRIDE = %s\n' % minimize_job_override)

      view_restrictions = info.get('view_restrictions')
      if view_restrictions:
        if view_restrictions in ALLOWED_VIEW_RESTRICTIONS:
          job.environment_string += (
              'ISSUE_VIEW_RESTRICTIONS = %s\n' % view_restrictions)
        else:
          logs.log_error('Invalid view restriction setting %s for project %s.' %
                         (view_restrictions, project))

      selective_unpack = info.get('selective_unpack')
      if selective_unpack:
        job.environment_string += 'UNPACK_ALL_FUZZ_TARGETS_AND_FILES = False\n'

      main_repo = info.get('main_repo')
      if main_repo:
        job.environment_string += f'MAIN_REPO = {main_repo}\n'

      if (template.engine == 'libfuzzer' and
          template.architecture == 'x86_64' and
          'dataflow' in info.get('fuzzing_engines', DEFAULT_ENGINES)):
        # Dataflow binaries are built with dataflow sanitizer, but can be used
        # as an auxiliary build with libFuzzer builds (e.g. with ASan or UBSan).
        dataflow_build_bucket_path = self._get_build_bucket_path(
            project_name=project,
            info=info,
            engine='dataflow',
            memory_tool='dataflow',
            architecture=template.architecture)
        job.environment_string += (
            'DATAFLOW_BUILD_BUCKET_PATH = %s\n' % dataflow_build_bucket_path)

      if self._additional_vars:
        additional_vars = {}
        additional_vars.update(self._additional_vars.get('all', {}))

        engine_vars = self._additional_vars.get(template.engine, {})
        engine_sanitizer_vars = engine_vars.get(template.memory_tool, {})
        additional_vars.update(engine_sanitizer_vars)

        for key, value in sorted(six.iteritems(additional_vars)):
          job.environment_string += ('{} = {}\n'.format(
              key,
              str(value).encode('unicode-escape').decode('utf-8')))

      job.put()

    return job_names

  def sync_user_permissions(self, project, info):
    """Sync permissions of project based on project.yaml."""
    ccs = ccs_from_info(info)

    for template in get_jobs_for_project(project, info):
      job_name = template.job_name(project, self._config_suffix)

      # Delete removed CCs.
      existing_ccs = data_types.ExternalUserPermission.query(
          data_types.ExternalUserPermission.entity_kind ==
          data_types.PermissionEntityKind.JOB,
          data_types.ExternalUserPermission.entity_name == job_name)
      ndb_utils.delete_multi([
          permission.key
          for permission in existing_ccs
          if permission.email not in ccs
      ])

      for cc in ccs:
        query = data_types.ExternalUserPermission.query(
            data_types.ExternalUserPermission.email == cc,
            data_types.ExternalUserPermission.entity_kind ==
            data_types.PermissionEntityKind.JOB,
            data_types.ExternalUserPermission.entity_name == job_name)

        existing_permission = query.get()
        if existing_permission:
          continue

        data_types.ExternalUserPermission(
            email=cc,
            entity_kind=data_types.PermissionEntityKind.JOB,
            entity_name=job_name,
            is_prefix=False,
            auto_cc=data_types.AutoCCType.ALL).put()

  def set_up(self, projects):
    """Do project setup. Return a list of all the project names that were set
    up."""
    job_names = []
    for project, info in projects:
      logs.log('Syncing configs for %s.' % project)

      backup_bucket_name = None
      corpus_bucket_name = None
      logs_bucket_name = None
      quarantine_bucket_name = None

      if self._segregate_projects:
        # Create per project service account and GCS buckets.
        (service_account, backup_bucket_name, corpus_bucket_name,
         logs_bucket_name, quarantine_bucket_name) = (
             self._create_service_accounts_and_buckets(project, info))

      # Create CF jobs for project.
      current_job_names = self._sync_job(project, info, corpus_bucket_name,
                                         quarantine_bucket_name,
                                         logs_bucket_name, backup_bucket_name)
      job_names.extend(current_job_names)

      if self._segregate_projects:
        self.sync_user_permissions(project, info)

        # Create Pub/Sub topics for tasks.
        create_pubsub_topics(project)

        # Set up projects settings (such as CPU distribution settings).
        if not info.get('disabled', False):
          create_project_settings(project, info, service_account)

    # Delete old/disabled project settings.
    enabled_projects = [
        project for project, info in projects
        if not info.get('disabled', False)
    ]
    return SetupResult(enabled_projects, job_names)


def cleanup_stale_projects(fuzzer_entities, project_names, job_names,
                           segregate_projects):
  """Clean up stale projects."""
  update_fuzzer_jobs(fuzzer_entities, job_names)
  cleanup_old_projects_settings(project_names)

  if segregate_projects:
    cleanup_pubsub_topics(project_names)


class Handler(base_handler.Handler):
  """Setup ClusterFuzz jobs for projects."""

  @handler.cron()
  def get(self):
    """Handles a GET request."""
    libfuzzer = data_types.Fuzzer.query(
        data_types.Fuzzer.name == 'libFuzzer').get()
    if not libfuzzer:
      logs.log_error('Failed to get libFuzzer Fuzzer entity.')
      return

    afl = data_types.Fuzzer.query(data_types.Fuzzer.name == 'afl').get()
    if not afl:
      logs.log_error('Failed to get AFL Fuzzer entity.')
      return

    honggfuzz = data_types.Fuzzer.query(
        data_types.Fuzzer.name == 'honggfuzz').get()
    if not honggfuzz:
      logs.log_error('Failed to get honggfuzz Fuzzer entity.')
      return

    gft = data_types.Fuzzer.query(
        data_types.Fuzzer.name == 'googlefuzztest').get()
    if not gft:
      logs.log_error('Failed to get googlefuzztest Fuzzer entity.')
      return

    project_config = local_config.ProjectConfig()
    segregate_projects = project_config.get('segregate_projects')
    project_setup_configs = project_config.get('project_setup')
    project_names = set()
    job_names = set()

    fuzzer_entities = {
        'afl': afl,
        'honggfuzz': honggfuzz,
        'googlefuzztest': gft,
        'libfuzzer': libfuzzer,
    }

    for setup_config in project_setup_configs:
      bucket_config = setup_config.get('build_buckets')

      if not bucket_config:
        raise ProjectSetupError('Project setup buckets not specified.')

      config = ProjectSetup(
          BUILD_BUCKET_PATH_TEMPLATE,
          REVISION_URL,
          setup_config.get('build_type'),
          config_suffix=setup_config.get('job_suffix', ''),
          external_config=setup_config.get('external_config', ''),
          segregate_projects=segregate_projects,
          experimental_sanitizers=setup_config.get('experimental_sanitizers',
                                                   []),
          engine_build_buckets={
              'libfuzzer': bucket_config.get('libfuzzer'),
              'libfuzzer-i386': bucket_config.get('libfuzzer_i386'),
              'afl': bucket_config.get('afl'),
              'honggfuzz': bucket_config.get('honggfuzz'),
              'googlefuzztest': bucket_config.get('googlefuzztest'),
              'none': bucket_config.get('no_engine'),
              'dataflow': bucket_config.get('dataflow'),
          },
          fuzzer_entities=fuzzer_entities,
          add_info_labels=setup_config.get('add_info_labels', False),
          add_revision_mappings=setup_config.get('add_revision_mappings',
                                                 False),
          additional_vars=setup_config.get('additional_vars'))

      projects_source = setup_config.get('source')
      if projects_source == 'oss-fuzz':
        projects = get_oss_fuzz_projects()
      elif projects_source.startswith(storage.GS_PREFIX):
        projects = get_projects_from_gcs(projects_source)
      else:
        raise ProjectSetupError('Invalid projects source: ' + projects_source)

      if not projects:
        raise ProjectSetupError('Missing projects list.')

      result = config.set_up(projects)
      project_names.update(result.project_names)
      job_names.update(result.job_names)

    cleanup_stale_projects(
        list(fuzzer_entities.values()), project_names, job_names,
        segregate_projects)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""AFL engine interface."""
import os
import stat

from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.bot.fuzzers.afl import launcher
from clusterfuzz._internal.bot.fuzzers.afl import stats
from clusterfuzz._internal.metrics import logs
from clusterfuzz.fuzz import engine


def _run_single_testcase(fuzzer_runner, testcase_file_path):
  """Loads a crash testcase if it exists."""
  # To ensure that we can run the fuzzer.
  os.chmod(fuzzer_runner.executable_path, stat.S_IRWXU | stat.S_IRGRP
           | stat.S_IXGRP)

  return fuzzer_runner.run_single_testcase(testcase_file_path)


class AFLEngine(engine.Engine):
  """AFL engine implementation."""

  @property
  def name(self):
    return 'afl'

  def prepare(self, corpus_dir, target_path, build_dir):  # pylint: disable=unused-argument
    """Prepare for a fuzzing session, by generating options.

    Returns a FuzzOptions object.

    Args:
      corpus_dir: The main corpus directory.
      target_path: Path to the target.
      build_dir: Path to the build directory.

    Returns:
      A FuzzOptions object.
    """
    afl_config = launcher.AflConfig.from_target_path(target_path)
    arguments = afl_config.additional_afl_arguments
    # TODO(mbarbella): Select all strategies here instead of deferring to fuzz.

    if self.do_strategies:
      strategies = launcher.FuzzingStrategies(target_path).to_strategy_dict()
    else:
      strategies = {}

    return engine.FuzzOptions(corpus_dir, arguments, strategies)

  def fuzz(self, target_path, options, reproducers_dir, max_time):
    """Run a fuzz session.

    Args:
      target_path: Path to the target.
      options: The FuzzOptions object returned by prepare().
      reproducers_dir: The directory to put reproducers in when crashes are
        found.
      max_time: Maximum allowed time for the fuzzing to run.

   Returns:
      A FuzzResult object.
    """
    config = launcher.AflConfig.from_target_path(target_path)
    config.additional_afl_arguments = options.arguments

    testcase_file_path = os.path.join(reproducers_dir, 'crash')
    runner = launcher.prepare_runner(
        target_path,
        config,
        testcase_file_path,
        options.corpus_dir,
        max_time,
        strategy_dict=options.strategies)

    fuzz_result = runner.fuzz()

    command = fuzz_result.command
    time_executed = fuzz_result.time_executed
    fuzzing_logs = fuzz_result.output + runner.fuzzer_stderr

    # Bail out if AFL returns a nonzero status code.
    if fuzz_result.return_code:
      target = engine_common.get_project_qualified_fuzzer_name(target_path)
      logs.log_error(
          f'afl: engine encountered an error (target={target})',
          engine_output=fuzz_result.output)
      return engine.FuzzResult(fuzzing_logs, command, [], {}, time_executed)

    stats_getter = stats.StatsGetter(runner.afl_output.stats_path,
                                     config.dict_path)
    new_units_generated, new_units_added, corpus_size = (
        runner.libfuzzerize_corpus())
    stats_getter.set_stats(fuzz_result.time_executed, new_units_generated,
                           new_units_added, corpus_size, runner.strategies,
                           runner.fuzzer_stderr, fuzz_result.output)

    crashes = []
    if os.path.exists(testcase_file_path):
      crash = engine.Crash(testcase_file_path, runner.fuzzer_stderr, [],
                           fuzz_result.time_executed)
      crashes.append(crash)

    return engine.FuzzResult(fuzzing_logs, command, crashes, stats_getter.stats,
                             time_executed)

  def reproduce(self, target_path, input_path, arguments, max_time):
    """Reproduce a crash given an input.

    Args:
      target_path: Path to the target.
      input_path: Path to the reproducer input.
      arguments: Additional arguments needed for reproduction.
      max_time: Maximum allowed time for the reproduction.

    Returns:
      A ReproduceResult.
    """
    del arguments
    del max_time
    config = launcher.AflConfig.from_target_path(target_path)
    input_directory = None  # Not required for reproduction.
    runner = launcher.prepare_runner(target_path, config, input_path,
                                     input_directory)

    reproduce_result = _run_single_testcase(runner, input_path)

    command = reproduce_result.command
    return_code = reproduce_result.return_code
    time_executed = reproduce_result.time_executed
    output = runner.fuzzer_stderr

    return engine.ReproduceResult(command, return_code, time_executed, output)

  def minimize_corpus(self, target_path, arguments, input_dirs, output_dir,
                      reproducers_dir, max_time):
    """Optional (but recommended): run corpus minimization.

    Args:
      target_path: Path to the target.
      arguments: Additional arguments needed for corpus minimization.
      input_dirs: Input corpora.
      output_dir: Output directory to place minimized corpus.
      reproducers_dir: The directory to put reproducers in when crashes are
        found.
      max_time: Maximum allowed time for the minimization.

    Returns:
      A FuzzResult object.
    """
    # TODO(mbarbella): Implement this.
    raise NotImplementedError



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Corpus pruning task."""

import collections
import datetime
from enum import Enum
import os
import random
import shutil

from google.cloud import ndb

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.bot.fuzzers import options
from clusterfuzz._internal.bot.fuzzers.libFuzzer import constants
from clusterfuzz._internal.bot.tasks import setup
from clusterfuzz._internal.bot.tasks import task_creation
from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.crash_analysis import crash_analyzer
from clusterfuzz._internal.crash_analysis.stack_parsing import stack_analyzer
from clusterfuzz._internal.crash_analysis.stack_parsing import stack_symbolizer
from clusterfuzz._internal.datastore import corpus_tagging
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import fuzz_target_utils
from clusterfuzz._internal.fuzzing import corpus_manager
from clusterfuzz._internal.fuzzing import leak_blacklist
from clusterfuzz._internal.google_cloud_utils import big_query
from clusterfuzz._internal.google_cloud_utils import blobs
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import archive
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell
from clusterfuzz.fuzz import engine

# TODO(ochang): Move common libFuzzer code from fuzzer into CF.

# Redzone size for running testcase.
DEFAULT_REDZONE = 32

# Minimum redzone size for use during merging.
MIN_REDZONE = 16

# Timeout for corpus pruning. Note that our priority is to make sure fuzzer's
# own corpus is minimized. If time remains, we spend time on using units from
# shared corpus. This combined timeout should be lower than task lease timeout.
CORPUS_PRUNING_TIMEOUT = 22 * 60 * 60

# Time to allow libFuzzer to timeout on its own.
SINGLE_UNIT_TIMEOUT = 5
TIMEOUT_FLAG = '-timeout=%d' % SINGLE_UNIT_TIMEOUT

# Corpus files limit for cases when corpus pruning task failed in the last
# execution.
CORPUS_FILES_LIMIT_FOR_FAILURES = 10000

# Corpus total size limit for cases when corpus pruning task failed in the last
# execution.
CORPUS_SIZE_LIMIT_FOR_FAILURES = 2 * 1024 * 1024 * 1024  # 2 GB.

# Maximum number of units to restore from quarantine in one run.
MAX_QUARANTINE_UNITS_TO_RESTORE = 128

# Memory limits for testcase.
RSS_LIMIT = 2560
RSS_LIMIT_MB_FLAG = '-rss_limit_mb=%d'

# Flag to enforce length limit for a single corpus element.
MAX_LEN_FLAG = '-max_len=%d'

# Flag to control memory leaks detection.
DETECT_LEAKS_FLAG = '-detect_leaks=%d'

# Longer than default sync timeout to fix broken (overly large) corpora without
# losing coverage.
SYNC_TIMEOUT = 2 * 60 * 60

# Number of fuzz targets whose backup corpus is used to cross pollinate with our
# current fuzz target corpus.
CROSS_POLLINATE_FUZZER_COUNT = 3

CorpusPruningResult = collections.namedtuple('CorpusPruningResult', [
    'coverage_info', 'crashes', 'fuzzer_binary_name', 'revision',
    'cross_pollination_stats'
])

CorpusCrash = collections.namedtuple('CorpusCrash', [
    'crash_state',
    'crash_type',
    'crash_address',
    'crash_stacktrace',
    'unit_path',
    'security_flag',
])

CrossPollinationStats = collections.namedtuple('CrossPollinationStats', [
    'project_qualified_name', 'method', 'sources', 'tags',
    'initial_corpus_size', 'corpus_size', 'initial_edge_coverage',
    'edge_coverage', 'initial_feature_coverage', 'feature_coverage'
])


class Pollination(str, Enum):
  RANDOM = 'random'
  TAGGED = 'tagged'


def _get_corpus_file_paths(corpus_path):
  """Return full paths to corpus files in |corpus_path|."""
  return [
      os.path.join(corpus_path, filename)
      for filename in os.listdir(corpus_path)
  ]


def _limit_corpus_size(corpus_url):
  """Limit number of files and size of a corpus."""
  corpus_count = 0
  corpus_size = 0
  deleted_corpus_count = 0
  bucket, _ = storage.get_bucket_name_and_path(corpus_url)
  for corpus_file in storage.get_blobs(corpus_url):
    corpus_count += 1
    corpus_size += corpus_file['size']
    if (corpus_count > CORPUS_FILES_LIMIT_FOR_FAILURES or
        corpus_size > CORPUS_SIZE_LIMIT_FOR_FAILURES):
      path_to_delete = storage.get_cloud_storage_file_path(
          bucket, corpus_file['name'])
      storage.delete(path_to_delete)
      deleted_corpus_count += 1

  if deleted_corpus_count:
    logs.log('Removed %d files from oversized corpus: %s.' %
             (deleted_corpus_count, corpus_url))


def _get_time_remaining(start_time):
  """Return time remaining."""
  time_used = int((datetime.datetime.utcnow() - start_time).total_seconds())
  return CORPUS_PRUNING_TIMEOUT - time_used


class CorpusPruningException(Exception):
  """Corpus pruning exception."""


class CrossPollinateFuzzer(object):
  """Cross Pollinate Fuzzer."""

  def __init__(self, fuzz_target, backup_bucket_name, corpus_engine_name):
    self.fuzz_target = fuzz_target
    self.backup_bucket_name = backup_bucket_name
    self.corpus_engine_name = corpus_engine_name


class Context(object):
  """Pruning state."""

  def __init__(self,
               fuzz_target,
               cross_pollinate_fuzzers,
               cross_pollination_method=Pollination.RANDOM,
               tag=None):

    self.fuzz_target = fuzz_target
    self.cross_pollinate_fuzzers = cross_pollinate_fuzzers
    self.cross_pollination_method = cross_pollination_method
    self.tag = tag

    self.merge_tmp_dir = None
    self.engine = engine.get(self.fuzz_target.engine)
    if not self.engine:
      raise CorpusPruningException('Engine {} not found'.format(engine))

    self._created_directories = []

    # Set up temporary directories where corpora will be synced to.
    # Initial synced corpus.
    self.initial_corpus_path = self._create_temp_corpus_directory(
        '%s_initial_corpus' % self.fuzz_target.project_qualified_name())
    # Minimized corpus.
    self.minimized_corpus_path = self._create_temp_corpus_directory(
        '%s_minimized_corpus' % self.fuzz_target.project_qualified_name())
    # Synced quarantine corpus.
    self.quarantine_corpus_path = self._create_temp_corpus_directory(
        '%s_quarantine' % self.fuzz_target.project_qualified_name())
    # Synced shared corpus.
    self.shared_corpus_path = self._create_temp_corpus_directory(
        '%s_shared' % self.fuzz_target.project_qualified_name())
    # Bad units.
    self.bad_units_path = self._create_temp_corpus_directory(
        '%s_bad_units' % self.fuzz_target.project_qualified_name())
    self.merge_tmp_dir = self._create_temp_corpus_directory('merge_workdir')

    self.corpus = corpus_manager.FuzzTargetCorpus(
        self.fuzz_target.engine,
        self.fuzz_target.project_qualified_name(),
        include_regressions=True)
    self.quarantine_corpus = corpus_manager.FuzzTargetCorpus(
        self.fuzz_target.engine,
        self.fuzz_target.project_qualified_name(),
        quarantine=True)

    shared_corpus_bucket = environment.get_value('SHARED_CORPUS_BUCKET')
    self.shared_corpus = corpus_manager.GcsCorpus(shared_corpus_bucket)

  def restore_quarantined_units(self):
    """Restore units from the quarantine."""
    logs.log('Restoring units from quarantine.')
    # Limit the number of quarantine units to restore, in case there are a lot.
    quarantine_unit_paths = _get_corpus_file_paths(self.quarantine_corpus_path)
    if len(quarantine_unit_paths) > MAX_QUARANTINE_UNITS_TO_RESTORE:
      quarantine_unit_paths = random.sample(quarantine_unit_paths,
                                            MAX_QUARANTINE_UNITS_TO_RESTORE)

    for unit_path in quarantine_unit_paths:
      unit_filename = os.path.basename(unit_path)
      shutil.move(unit_path,
                  os.path.join(self.initial_corpus_path, unit_filename))

  def _create_temp_corpus_directory(self, name):
    """Create temporary corpus directory. Returns path to the created
    directory."""
    testcases_directory = environment.get_value('FUZZ_INPUTS_DISK')
    directory_path = os.path.join(testcases_directory, name)
    shell.create_directory(directory_path)
    self._created_directories.append(directory_path)

    return directory_path

  def sync_to_disk(self):
    """Sync required corpora to disk."""
    if not self.corpus.rsync_to_disk(
        self.initial_corpus_path, timeout=SYNC_TIMEOUT):
      raise CorpusPruningException('Failed to sync corpus to disk.')

    if not self.quarantine_corpus.rsync_to_disk(self.quarantine_corpus_path):
      logs.log_error(
          'Failed to sync quarantine corpus to disk.',
          fuzz_target=self.fuzz_target)

    if not self.shared_corpus.rsync_to_disk(self.shared_corpus_path):
      logs.log_error(
          'Failed to sync shared corpus to disk.', fuzz_target=self.fuzz_target)

    self._cross_pollinate_other_fuzzer_corpuses()

  def sync_to_gcs(self):
    """Sync corpora to GCS post merge."""
    if not self.corpus.rsync_from_disk(self.minimized_corpus_path):
      raise CorpusPruningException('Failed to sync minimized corpus to gcs.')

  def cleanup(self):
    """Cleanup state."""
    for path in self._created_directories:
      shell.remove_directory(path)

  def _cross_pollinate_other_fuzzer_corpuses(self):
    """Add other fuzzer corpuses to shared corpus path for cross-pollination."""
    corpus_backup_date = utils.utcnow().date() - datetime.timedelta(
        days=data_types.CORPUS_BACKUP_PUBLIC_LOOKBACK_DAYS)

    for cross_pollinate_fuzzer in self.cross_pollinate_fuzzers:
      project_qualified_name = (
          cross_pollinate_fuzzer.fuzz_target.project_qualified_name())
      backup_bucket_name = cross_pollinate_fuzzer.backup_bucket_name
      corpus_engine_name = cross_pollinate_fuzzer.corpus_engine_name

      corpus_backup_url = corpus_manager.gcs_url_for_backup_file(
          backup_bucket_name, corpus_engine_name, project_qualified_name,
          corpus_backup_date)
      corpus_backup_local_filename = '%s-%s' % (
          project_qualified_name, os.path.basename(corpus_backup_url))
      corpus_backup_local_path = os.path.join(self.shared_corpus_path,
                                              corpus_backup_local_filename)

      if not storage.exists(corpus_backup_url, ignore_errors=True):
        # This can happen in cases when a new fuzz target is checked in or if
        # missed to capture a backup for a particular day (for OSS-Fuzz, this
        # will result in a 403 instead of 404 since that GCS path belongs to
        # other project). So, just log a warning for debugging purposes only.
        logs.log_warn(
            'Corpus backup does not exist, ignoring: %s.' % corpus_backup_url)
        continue

      if not storage.copy_file_from(corpus_backup_url,
                                    corpus_backup_local_path):
        continue

      corpus_backup_output_directory = os.path.join(self.shared_corpus_path,
                                                    project_qualified_name)
      shell.create_directory(corpus_backup_output_directory)
      result = archive.unpack(corpus_backup_local_path,
                              corpus_backup_output_directory)
      shell.remove_file(corpus_backup_local_path)

      if result:
        logs.log(
            'Corpus backup url %s successfully unpacked into shared corpus.' %
            corpus_backup_url)
      else:
        logs.log_error(
            'Failed to unpack corpus backup from url %s.' % corpus_backup_url)


class Runner(object):
  """Runner for libFuzzer."""

  def __init__(self, build_directory, context):
    self.build_directory = build_directory
    self.context = context

    self.target_path = engine_common.find_fuzzer_path(
        self.build_directory, self.context.fuzz_target.binary)
    if not self.target_path:
      raise CorpusPruningException(
          'Failed to get fuzzer path for %s.' % self.context.fuzz_target.binary)

    self.fuzzer_options = options.get_fuzz_target_options(self.target_path)

  def get_libfuzzer_flags(self):
    """Get default libFuzzer options."""
    rss_limit = RSS_LIMIT
    max_len = engine_common.CORPUS_INPUT_SIZE_LIMIT
    detect_leaks = 1
    arguments = [TIMEOUT_FLAG]

    if self.fuzzer_options:
      # Default values from above can be customized for a given fuzz target.
      libfuzzer_arguments = self.fuzzer_options.get_engine_arguments(
          'libfuzzer')

      custom_rss_limit = libfuzzer_arguments.get(
          'rss_limit_mb', constructor=int)
      if custom_rss_limit and custom_rss_limit < rss_limit:
        rss_limit = custom_rss_limit

      custom_max_len = libfuzzer_arguments.get('max_len', constructor=int)
      if custom_max_len and custom_max_len < max_len:
        max_len = custom_max_len

      # Some targets might falsely report leaks all the time, so allow this to
      # be disabled.
      custom_detect_leaks = libfuzzer_arguments.get(
          'detect_leaks', constructor=int)
      if custom_detect_leaks is not None:
        detect_leaks = custom_detect_leaks

    arguments.append(RSS_LIMIT_MB_FLAG % rss_limit)
    arguments.append(MAX_LEN_FLAG % max_len)
    arguments.append(DETECT_LEAKS_FLAG % detect_leaks)
    arguments.append(constants.VALUE_PROFILE_ARGUMENT)

    return arguments

  def process_sanitizer_options(self):
    """Process sanitizer options overrides."""
    if not self.fuzzer_options:
      return

    # Only need to look as ASan, as that's what we prune with.
    overrides = self.fuzzer_options.get_asan_options()
    if not overrides:
      return

    asan_options = environment.get_memory_tool_options('ASAN_OPTIONS')
    asan_options.update(overrides)
    environment.set_memory_tool_options('ASAN_OPTIONS', asan_options)

  def reproduce(self, input_path, arguments, max_time):
    return self.context.engine.reproduce(self.target_path, input_path,
                                         arguments, max_time)

  def minimize_corpus(self, arguments, input_dirs, output_dir, reproducers_dir,
                      max_time):
    return self.context.engine.minimize_corpus(self.target_path, arguments,
                                               input_dirs, output_dir,
                                               reproducers_dir, max_time)


class CorpusPruner(object):
  """Class that handles corpus pruning."""

  def __init__(self, runner):
    self.runner = runner
    self.context = self.runner.context

  def _run_single_unit(self, unit_path):
    """Run a single unit, and return the result."""
    arguments = self.runner.get_libfuzzer_flags()
    return self.runner.reproduce(unit_path, arguments, SINGLE_UNIT_TIMEOUT)

  def _quarantine_unit(self, unit_path, quarantine_corpus_path):
    """Moves the given unit to the quarantine, and returns the path to the unit
    in the quarantine."""
    quarantined_unit_path = os.path.join(quarantine_corpus_path,
                                         os.path.basename(unit_path))
    shutil.move(unit_path, quarantined_unit_path)

    return quarantined_unit_path

  def process_bad_units(self, bad_units_path, quarantine_corpus_path, crashes):
    """Process bad units found during merge."""
    # TODO(ochang): A lot of this function is similar to parts of fuzz_task.
    # Ideally fuzz_task can be refactored in a way that lets us share the common
    # code.

    environment.reset_current_memory_tool_options(redzone_size=DEFAULT_REDZONE)
    self.runner.process_sanitizer_options()

    logs.log('Processing bad units.')
    corpus_file_paths = _get_corpus_file_paths(bad_units_path)
    num_bad_units = 0

    # Run each corpus item individually.
    for i, unit_path in enumerate(corpus_file_paths, 1):
      if i % 100 == 0:
        logs.log('Up to %d' % i)

      unit_name = os.path.basename(unit_path)
      if unit_name.startswith('timeout-') or unit_name.startswith('oom-'):
        # Don't waste time re-running timeout or oom testcases.
        self._quarantine_unit(unit_path, quarantine_corpus_path)
        num_bad_units += 1
        continue

      try:
        result = self._run_single_unit(unit_path)
      except TimeoutError:
        # Slow unit. Quarantine it.
        self._quarantine_unit(unit_path, quarantine_corpus_path)
        num_bad_units += 1
        continue

      if not crash_analyzer.is_memory_tool_crash(result.output):
        # Didn't crash.
        continue

      # Get memory tool crash information.
      state = stack_analyzer.get_crash_data(result.output, symbolize_flag=True)

      # Crashed or caused a leak. Quarantine it.
      unit_path = self._quarantine_unit(unit_path, quarantine_corpus_path)
      num_bad_units += 1

      if crash_analyzer.ignore_stacktrace(state.crash_stacktrace):
        continue

      # Local de-duplication.
      if state.crash_state not in crashes:
        security_flag = crash_analyzer.is_security_issue(
            state.crash_stacktrace, state.crash_type, state.crash_address)
        crashes[state.crash_state] = CorpusCrash(
            state.crash_state, state.crash_type, state.crash_address,
            state.crash_stacktrace, unit_path, security_flag)

    logs.log('Found %d bad units, %d unique crashes.' % (num_bad_units,
                                                         len(crashes)))

  def run(self, initial_corpus_path, minimized_corpus_path, bad_units_path):
    """Run corpus pruning. Output result to directory."""
    if not shell.get_directory_file_count(initial_corpus_path):
      # Empty corpus, nothing to do.
      return None

    # Set memory tool options and fuzzer arguments.
    engine_common.unpack_seed_corpus_if_needed(
        self.runner.target_path, initial_corpus_path, force_unpack=True)

    environment.reset_current_memory_tool_options(
        redzone_size=MIN_REDZONE, leaks=True)
    self.runner.process_sanitizer_options()
    additional_args = self.runner.get_libfuzzer_flags()

    # Execute fuzzer with arguments for corpus pruning.
    logs.log('Running merge...')
    try:
      result = self.runner.minimize_corpus(
          additional_args, [initial_corpus_path], minimized_corpus_path,
          bad_units_path, CORPUS_PRUNING_TIMEOUT)
    except TimeoutError as e:
      raise CorpusPruningException(
          'Corpus pruning timed out while minimizing corpus\n' + repr(e))
    except engine.Error as e:
      raise CorpusPruningException(
          'Corpus pruning failed to minimize corpus\n' + repr(e))

    symbolized_output = stack_symbolizer.symbolize_stacktrace(result.logs)

    # Sanity check that there are files in minimized corpus after merging.
    if not shell.get_directory_file_count(minimized_corpus_path):
      raise CorpusPruningException(
          'Corpus pruning failed to minimize corpus\n' + symbolized_output)

    logs.log('Corpus merge finished successfully.', output=symbolized_output)

    return result.stats


class CrossPollinator(object):
  """Cross pollination."""

  def __init__(self, runner):
    self.runner = runner
    self.context = self.runner.context

  def run(self, timeout):
    """Merge testcases from corpus from other fuzz targets."""
    if not shell.get_directory_file_count(self.context.shared_corpus_path):
      logs.log('No files found in shared corpus, skip merge.')
      return None

    # Run pruning on the shared corpus and log the result in case of error.
    logs.log('Merging shared corpus...')
    environment.reset_current_memory_tool_options(redzone_size=DEFAULT_REDZONE)
    self.runner.process_sanitizer_options()

    additional_args = self.runner.get_libfuzzer_flags()

    try:
      result = self.runner.minimize_corpus(additional_args,
                                           [self.context.shared_corpus_path],
                                           self.context.minimized_corpus_path,
                                           self.context.bad_units_path, timeout)
      symbolized_output = stack_symbolizer.symbolize_stacktrace(result.logs)
      logs.log(
          'Shared corpus merge finished successfully.',
          output=symbolized_output)
    except TimeoutError as e:
      # Other cross pollinated fuzzer corpuses can have unexpected test cases
      # that time us out. This is expected, so bail out.
      logs.log_warn('Corpus pruning timed out while merging shared corpus\n' +
                    repr(e))
      return None
    except engine.Error as e:
      # Other cross pollinated fuzzer corpuses can be large, so we can run out
      # of disk space and exception out. This is expected, so bail out.
      logs.log_warn('Corpus pruning failed to merge shared corpus\n' + repr(e))
      return None

    return result.stats


def _record_cross_pollination_stats(stats):
  """Log stats about cross pollination in BigQuery."""
  # If no stats were gathered due to a timeout or lack of corpus, return.
  if not stats:
    return

  bigquery_row = {
      'project_qualified_name': stats.project_qualified_name,
      'method': stats.method,
      'sources': stats.sources,
      'tags': stats.tags if stats.tags else '',
      'initial_corpus_size': stats.initial_corpus_size,
      'corpus_size': stats.corpus_size,
      'initial_edge_coverage': stats.initial_edge_coverage,
      'edge_coverage': stats.edge_coverage,
      'initial_feature_coverage': stats.initial_feature_coverage,
      'feature_coverage': stats.feature_coverage
  }

  # BigQuery not available in local development. This is necessary because the
  # untrusted runner is in a separate process and can't be easily mocked.
  # Check here instead of earlier to test as much of the function as we can.
  if environment.get_value('LOCAL_DEVELOPMENT') or environment.get_value(
      'PY_UNITTESTS'):
    return

  client = big_query.Client(
      dataset_id='main', table_id='cross_pollination_statistics')
  client.insert([big_query.Insert(row=bigquery_row, insert_id=None)])


def do_corpus_pruning(context, last_execution_failed, revision):
  """Run corpus pruning."""
  # Set |FUZZ_TARGET| environment variable to help with unarchiving only fuzz
  # target and its related files.
  environment.set_value('FUZZ_TARGET', context.fuzz_target.binary)

  if environment.is_trusted_host():
    from clusterfuzz._internal.bot.untrusted_runner import tasks_host
    return tasks_host.do_corpus_pruning(context, last_execution_failed,
                                        revision)

  if not build_manager.setup_build(revision=revision):
    raise CorpusPruningException('Failed to setup build.')

  build_directory = environment.get_value('BUILD_DIR')
  start_time = datetime.datetime.utcnow()
  runner = Runner(build_directory, context)
  pruner = CorpusPruner(runner)
  fuzzer_binary_name = os.path.basename(runner.target_path)

  # If our last execution failed, shrink to a randomized corpus of usable size
  # to prevent corpus from growing unbounded and recurring failures when trying
  # to minimize it.
  if last_execution_failed:
    for corpus_url in [
        context.corpus.get_gcs_url(),
        context.quarantine_corpus.get_gcs_url()
    ]:
      _limit_corpus_size(corpus_url)

  # Get initial corpus to process from GCS.
  context.sync_to_disk()
  initial_corpus_size = shell.get_directory_file_count(
      context.initial_corpus_path)

  # Restore a small batch of quarantined units back to corpus.
  context.restore_quarantined_units()

  # Shrink to a minimized corpus using corpus merge.
  pruner_stats = pruner.run(context.initial_corpus_path,
                            context.minimized_corpus_path,
                            context.bad_units_path)

  # Sync minimized corpus back to GCS.
  context.sync_to_gcs()

  # Create corpus backup.
  # Temporarily copy the past crash regressions folder into the minimized corpus
  # so that corpus backup archive can have both.
  regressions_input_dir = os.path.join(context.initial_corpus_path,
                                       'regressions')
  regressions_output_dir = os.path.join(context.minimized_corpus_path,
                                        'regressions')
  if shell.get_directory_file_count(regressions_input_dir):
    shutil.copytree(regressions_input_dir, regressions_output_dir)
  backup_bucket = environment.get_value('BACKUP_BUCKET')
  corpus_backup_url = corpus_manager.backup_corpus(
      backup_bucket, context.corpus, context.minimized_corpus_path)
  shell.remove_directory(regressions_output_dir)

  minimized_corpus_size_units = shell.get_directory_file_count(
      context.minimized_corpus_path)
  minimized_corpus_size_bytes = shell.get_directory_size(
      context.minimized_corpus_path)

  logs.log('Corpus pruned from %d to %d units.' % (initial_corpus_size,
                                                   minimized_corpus_size_units))

  # Process bad units found during merge.
  # Mapping of crash state -> CorpusCrash
  crashes = {}
  pruner.process_bad_units(context.bad_units_path,
                           context.quarantine_corpus_path, crashes)
  context.quarantine_corpus.rsync_from_disk(context.quarantine_corpus_path)

  # Store corpus stats into CoverageInformation entity.
  project_qualified_name = context.fuzz_target.project_qualified_name()
  today = datetime.datetime.utcnow().date()
  coverage_info = data_types.CoverageInformation(
      fuzzer=project_qualified_name, date=today)

  quarantine_corpus_size = shell.get_directory_file_count(
      context.quarantine_corpus_path)
  quarantine_corpus_dir_size = shell.get_directory_size(
      context.quarantine_corpus_path)

  # Save the minimize corpus size before cross pollination to put in BigQuery.
  pre_pollination_corpus_size = minimized_corpus_size_units

  # Populate coverage stats.
  coverage_info.corpus_size_units = minimized_corpus_size_units
  coverage_info.corpus_size_bytes = minimized_corpus_size_bytes
  coverage_info.quarantine_size_units = quarantine_corpus_size
  coverage_info.quarantine_size_bytes = quarantine_corpus_dir_size
  coverage_info.corpus_backup_location = corpus_backup_url
  coverage_info.corpus_location = context.corpus.get_gcs_url()
  coverage_info.quarantine_location = context.quarantine_corpus.get_gcs_url()

  # Calculate remaining time to use for shared corpus merging.
  time_remaining = _get_time_remaining(start_time)
  if time_remaining <= 0:
    logs.log_warn('Not enough time for shared corpus merging.')
    return None

  cross_pollinator = CrossPollinator(runner)
  pollinator_stats = cross_pollinator.run(time_remaining)

  context.sync_to_gcs()

  # Update corpus size stats.
  minimized_corpus_size_units = shell.get_directory_file_count(
      context.minimized_corpus_path)
  minimized_corpus_size_bytes = shell.get_directory_size(
      context.minimized_corpus_path)
  coverage_info.corpus_size_units = minimized_corpus_size_units
  coverage_info.corpus_size_bytes = minimized_corpus_size_bytes

  logs.log('Finished.')

  sources = ','.join([
      fuzzer.fuzz_target.project_qualified_name()
      for fuzzer in context.cross_pollinate_fuzzers
  ])

  cross_pollination_stats = None
  if pruner_stats and pollinator_stats:
    cross_pollination_stats = CrossPollinationStats(
        project_qualified_name, context.cross_pollination_method, sources,
        context.tag, initial_corpus_size, pre_pollination_corpus_size,
        pruner_stats['edge_coverage'], pollinator_stats['edge_coverage'],
        pruner_stats['feature_coverage'], pollinator_stats['feature_coverage'])

  return CorpusPruningResult(
      coverage_info=coverage_info,
      crashes=list(crashes.values()),
      fuzzer_binary_name=fuzzer_binary_name,
      revision=environment.get_value('APP_REVISION'),
      cross_pollination_stats=cross_pollination_stats)


def _process_corpus_crashes(context, result):
  """Process crashes found in the corpus."""
  # Default Testcase entity values.
  crash_revision = result.revision
  job_type = environment.get_value('JOB_NAME')
  minimized_arguments = '%TESTCASE% ' + context.fuzz_target.binary
  project_name = data_handler.get_project_name(job_type)

  comment = 'Fuzzer %s generated corpus testcase crashed (r%s)' % (
      context.fuzz_target.project_qualified_name(), crash_revision)

  # Generate crash reports.
  for crash in result.crashes:
    existing_testcase = data_handler.find_testcase(
        project_name, crash.crash_type, crash.crash_state, crash.security_flag)
    if existing_testcase:
      continue

    # Upload/store testcase.
    if environment.is_trusted_host():
      from clusterfuzz._internal.bot.untrusted_runner import file_host
      unit_path = os.path.join(context.bad_units_path,
                               os.path.basename(crash.unit_path))
      # Prevent the worker from escaping out of |context.bad_units_path|.
      if not file_host.is_directory_parent(unit_path, context.bad_units_path):
        raise CorpusPruningException('Invalid units path from worker.')

      file_host.copy_file_from_worker(crash.unit_path, unit_path)
    else:
      unit_path = crash.unit_path

    with open(unit_path, 'rb') as f:
      key = blobs.write_blob(f)

    # Set the absolute_path property of the Testcase to a file in FUZZ_INPUTS
    # instead of the local quarantine directory.
    absolute_testcase_path = os.path.join(
        environment.get_value('FUZZ_INPUTS'), 'testcase')

    testcase_id = data_handler.store_testcase(
        crash=crash,
        fuzzed_keys=key,
        minimized_keys='',
        regression='',
        fixed='',
        one_time_crasher_flag=False,
        crash_revision=crash_revision,
        comment=comment,
        absolute_path=absolute_testcase_path,
        fuzzer_name=context.fuzz_target.engine,
        fully_qualified_fuzzer_name=context.fuzz_target.fully_qualified_name(),
        job_type=job_type,
        archived=False,
        archive_filename='',
        binary_flag=True,
        http_flag=False,
        gestures=None,
        redzone=DEFAULT_REDZONE,
        disable_ubsan=False,
        minidump_keys=None,
        window_argument=None,
        timeout_multiplier=1.0,
        minimized_arguments=minimized_arguments)

    # Set fuzzer_binary_name in testcase metadata.
    testcase = data_handler.get_testcase_by_id(testcase_id)
    testcase.set_metadata('fuzzer_binary_name', result.fuzzer_binary_name)

    issue_metadata = engine_common.get_all_issue_metadata_for_testcase(testcase)
    if issue_metadata:
      for key, value in issue_metadata.items():
        testcase.set_metadata(key, value, update_testcase=False)

      testcase.put()

    # Create additional tasks for testcase (starting with minimization).
    testcase = data_handler.get_testcase_by_id(testcase_id)
    task_creation.create_tasks(testcase)


def _select_targets_and_jobs_for_pollination(engine_name, current_fuzzer_name,
                                             method, tag):
  """Select jobs to use for cross pollination."""
  target_jobs = list(fuzz_target_utils.get_fuzz_target_jobs(engine=engine_name))

  if method == Pollination.TAGGED:
    similar_tagged_targets = [
        target.fully_qualified_fuzz_target_name
        for target in corpus_tagging.get_targets_with_tag(tag)
        if target.fully_qualified_fuzz_target_name != current_fuzzer_name
    ]
    # Intersect target_jobs and similar_tagged_targets on fully qualified
    # fuzz target name.
    target_jobs = [
        target for target in target_jobs
        if target.fuzz_target_name in similar_tagged_targets
    ]

  targets = fuzz_target_utils.get_fuzz_targets_for_target_jobs(target_jobs)

  targets_and_jobs = [(target, target_job)
                      for target, target_job in zip(targets, target_jobs)
                      if target_job.fuzz_target_name != current_fuzzer_name]
  selected_targets_and_jobs = random.SystemRandom().sample(
      targets_and_jobs, min(
          len(targets_and_jobs), CROSS_POLLINATE_FUZZER_COUNT))

  return selected_targets_and_jobs


def _get_cross_pollinate_fuzzers(engine_name, current_fuzzer_name, method, tag):
  """Return a list of fuzzer objects to use for cross pollination."""
  cross_pollinate_fuzzers = []

  selected_targets_and_jobs = _select_targets_and_jobs_for_pollination(
      engine_name, current_fuzzer_name, method, tag)

  default_backup_bucket = utils.default_backup_bucket()
  for target, target_job in selected_targets_and_jobs:
    job = data_types.Job.query(data_types.Job.name == target_job.job).get()
    if not job:
      continue

    job_environment = job.get_environment()
    backup_bucket_name = job_environment.get('BACKUP_BUCKET',
                                             default_backup_bucket)
    if not backup_bucket_name:
      continue
    corpus_engine_name = job_environment.get('CORPUS_FUZZER_NAME_OVERRIDE',
                                             engine_name)

    cross_pollinate_fuzzers.append(
        CrossPollinateFuzzer(
            target,
            backup_bucket_name,
            corpus_engine_name,
        ))

  return cross_pollinate_fuzzers


def _save_coverage_information(context, result):
  """Saves coverage information in datastore using an atomic transaction."""

  # Use ndb.transaction with retries below to mitigate risk of a race condition.
  def _try_save_coverage_information():
    """Implements save_coverage_information function."""
    coverage_info = data_handler.get_coverage_information(
        context.fuzz_target.project_qualified_name(),
        result.coverage_info.date,
        create_if_needed=True)

    # Intentionally skip edge and function coverage values as those would come
    # from fuzzer coverage cron task (see src/go/server/cron/coverage.go).
    coverage_info.corpus_size_units = result.coverage_info.corpus_size_units
    coverage_info.corpus_size_bytes = result.coverage_info.corpus_size_bytes
    coverage_info.corpus_location = result.coverage_info.corpus_location
    coverage_info.corpus_backup_location = (
        result.coverage_info.corpus_backup_location)
    coverage_info.quarantine_size_units = (
        result.coverage_info.quarantine_size_units)
    coverage_info.quarantine_size_bytes = (
        result.coverage_info.quarantine_size_bytes)
    coverage_info.quarantine_location = result.coverage_info.quarantine_location
    coverage_info.put()

  try:
    ndb.transaction(
        _try_save_coverage_information,
        retries=data_handler.DEFAULT_FAIL_RETRIES)
  except Exception as e:
    raise CorpusPruningException(
        'Failed to save corpus pruning result: %s.' % repr(e))


def choose_cross_pollination_strategy(current_fuzzer_name):
  """Chooses cross pollination strategy. In seperate function to mock for
    predictable test behaviror."""
  method = random.choice([Pollination.RANDOM, Pollination.TAGGED])
  if method == Pollination.TAGGED:
    similar_targets = corpus_tagging.get_similarly_tagged_fuzzers(
        current_fuzzer_name)
    if similar_targets:
      return (Pollination.TAGGED, random.choice(list(similar_targets.keys())))

  return (Pollination.RANDOM, None)


def execute_task(full_fuzzer_name, job_type):
  """Execute corpus pruning task."""
  fuzz_target = data_handler.get_fuzz_target(full_fuzzer_name)
  task_name = 'corpus_pruning_%s_%s' % (full_fuzzer_name, job_type)
  revision = 0  # Trunk revision

  # Get status of last execution.
  last_execution_metadata = data_handler.get_task_status(task_name)
  last_execution_failed = (
      last_execution_metadata and
      last_execution_metadata.status == data_types.TaskState.ERROR)

  # Make sure we're the only instance running for the given fuzzer and
  # job_type.
  if not data_handler.update_task_status(task_name,
                                         data_types.TaskState.STARTED):
    logs.log('A previous corpus pruning task is still running, exiting.')
    return

  # Setup fuzzer and data bundle.
  if not setup.update_fuzzer_and_data_bundles(fuzz_target.engine):
    raise CorpusPruningException(
        'Failed to set up fuzzer %s.' % fuzz_target.engine)

  cross_pollination_method, tag = choose_cross_pollination_strategy(
      full_fuzzer_name)

  # TODO(unassigned): Use coverage information for better selection here.
  cross_pollinate_fuzzers = _get_cross_pollinate_fuzzers(
      fuzz_target.engine, full_fuzzer_name, cross_pollination_method, tag)

  context = Context(fuzz_target, cross_pollinate_fuzzers,
                    cross_pollination_method, tag)

  # Copy global blacklist into local suppressions file if LSan is enabled.
  is_lsan_enabled = environment.get_value('LSAN')
  if is_lsan_enabled:
    # TODO(ochang): Copy this to untrusted worker.
    leak_blacklist.copy_global_to_local_blacklist()

  try:
    result = do_corpus_pruning(context, last_execution_failed, revision)
    _record_cross_pollination_stats(result.cross_pollination_stats)
    _save_coverage_information(context, result)
    _process_corpus_crashes(context, result)
  except Exception:
    logs.log_error('Corpus pruning failed.')
    data_handler.update_task_status(task_name, data_types.TaskState.ERROR)
    return
  finally:
    context.cleanup()

  data_handler.update_task_status(task_name, data_types.TaskState.FINISHED)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""External reproduction updates."""

from clusterfuzz._internal.crash_analysis import crash_analyzer
from clusterfuzz._internal.crash_analysis.crash_comparer import CrashComparer
from clusterfuzz._internal.crash_analysis.stack_parsing import stack_analyzer
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.metrics import logs
from handlers import base_handler
from libs import handler
from libs import helpers


def _mark_as_fixed(testcase, revision):
  """Mark bug as fixed."""
  testcase.open = False
  # Bisection not available for external reproduction infrastructure. Assume
  # range (crash revision : current revision).
  testcase.fixed = f'{testcase.crash_revision}:{revision}'
  data_handler.update_progression_completion_metadata(
      testcase, revision, message=f'fixed in r{revision}')


def _mark_errored(testcase, revision, error):
  """Mark testcase as errored out."""
  message = 'Received error from external infra, marking testcase as NA.'
  logs.log_warn(message, error=error, testcase_id=testcase.key.id())

  testcase.fixed = 'NA'
  testcase.open = False
  data_handler.update_progression_completion_metadata(
      testcase, revision, message=message)


def handle_update(testcase, revision, stacktrace, error):
  """Handle update."""
  logs.log('Got external update for testcase.', testcase_id=testcase.key.id())
  if error:
    _mark_errored(testcase, revision, error)
    return

  last_tested_revision = (
      testcase.get_metadata('last_tested_revision') or testcase.crash_revision)

  if revision < last_tested_revision:
    logs.log_warn(f'Revision {revision} less than previously tested '
                  f'revision {last_tested_revision}.')
    return

  fuzz_target = testcase.get_fuzz_target()
  if fuzz_target:
    fuzz_target_name = fuzz_target.binary
  else:
    fuzz_target_name = None

  # Record use of fuzz target to avoid garbage collection (since fuzz_task does
  # not run).
  data_handler.record_fuzz_target(fuzz_target.engine, fuzz_target.binary,
                                  testcase.job_type)

  state = stack_analyzer.get_crash_data(
      stacktrace,
      fuzz_target=fuzz_target_name,
      symbolize_flag=False,
      already_symbolized=True,
      detect_ooms_and_hangs=True)
  crash_comparer = CrashComparer(state.crash_state, testcase.crash_state)
  if not crash_comparer.is_similar():
    logs.log(f'State no longer similar ('
             f'testcase_id={testcase.key.id()}, '
             f'old_state={testcase.crash_state}, '
             f'new_state={state.crash_state})')
    _mark_as_fixed(testcase, revision)
    return

  is_security = crash_analyzer.is_security_issue(
      state.crash_stacktrace, state.crash_type, state.crash_address)
  if is_security != testcase.security_flag:
    logs.log(f'Security flag for {testcase.key.id()} no longer matches.')
    _mark_as_fixed(testcase, revision)
    return

  logs.log(f'{testcase.key.id()} still crashes.')
  testcase.last_tested_crash_stacktrace = stacktrace
  data_handler.update_progression_completion_metadata(
      testcase, revision, is_crash=True)


class Handler(base_handler.Handler):
  """External reproduction update."""

  @handler.pubsub_push
  def post(self, message):
    """Handle a post request."""
    testcase_id = message.attributes.get('testcaseId')
    if not testcase_id:
      raise helpers.EarlyExitException('Missing testcaseId.', 400)

    revision = message.attributes.get('revision')
    if not revision or not revision.isdigit():
      raise helpers.EarlyExitException('Missing revision.', 400)

    revision = int(revision)
    testcase = data_handler.get_testcase_by_id(testcase_id)
    job = data_types.Job.query(data_types.Job.name == testcase.job_type).get()
    if not job or not job.is_external():
      raise helpers.EarlyExitException('Invalid job.', 400)

    if message.data:
      stacktrace = message.data.decode()
    else:
      logs.log(f'No stacktrace provided (testcase_id={testcase_id}).')
      stacktrace = ''

    error = message.attributes.get('error')
    handle_update(testcase, revision, stacktrace, error)
    return 'OK'



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""json_utils tests."""

import datetime
import unittest

from clusterfuzz._internal.base import json_utils


class JSONDumpsTest(unittest.TestCase):
  """Test json_utils.dumps."""

  def test(self):
    """Tests json_utils.dumps with various primitive data types."""
    self.assertEqual(json_utils.dumps('string'), '"string"')

    self.assertEqual(json_utils.dumps(True), 'true')
    self.assertEqual(json_utils.dumps(False), 'false')

    self.assertEqual(json_utils.dumps(-5), '-5')
    self.assertEqual(json_utils.dumps(0), '0')
    self.assertEqual(json_utils.dumps(10), '10')
    self.assertEqual(json_utils.dumps(12.0), '12.0')

    self.assertEqual(
        json_utils.dumps(datetime.datetime(2018, 2, 4, 1, 2, 3, 9)),
        '{"__type__": "datetime", "day": 4, "hour": 1, "microsecond": 9, '
        '"minute": 2, "month": 2, "second": 3, "year": 2018}')
    self.assertEqual(
        json_utils.dumps(datetime.date(2018, 3, 20)),
        '{"__type__": "date", "day": 20, "month": 3, "year": 2018}')


class JSONLoadsTest(unittest.TestCase):
  """Test json_utils.loads."""

  def test(self):
    """Tests json_utils.loads with various primitive data types."""
    self.assertEqual(json_utils.loads('"string"'), 'string')

    self.assertEqual(json_utils.loads('true'), True)
    self.assertEqual(json_utils.loads('false'), False)

    self.assertEqual(json_utils.loads('-5'), -5)
    self.assertEqual(json_utils.loads('0'), 0)
    self.assertEqual(json_utils.loads('10'), 10)
    self.assertEqual(json_utils.loads('12.0'), 12.0)

    self.assertEqual(
        json_utils.loads(
            '{"second": 3, "microsecond": 9, "hour": 1, "year": 2018, '
            '"__type__": "datetime", "day": 4, "minute": 2, "month": 2}'),
        datetime.datetime(2018, 2, 4, 1, 2, 3, 9))

    self.assertEqual(
        json_utils.loads(
            '{"__type__": "date", "day": 20, "month": 3, "year": 2018}'),
        datetime.date(2018, 3, 20))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Metrics. All metrics should go into this file."""

from clusterfuzz._internal.metrics import monitor

# Fuzz task metrics.
BIG_QUERY_WRITE_COUNT = monitor.CounterMetric(
    'debug/big_query/write_count',
    description='The number of BigQuery writes',
    field_spec=[
        monitor.BooleanField('success'),
    ])

JOB_BAD_BUILD_COUNT = monitor.CounterMetric(
    'task/fuzz/job/bad_build_count',
    description=("Count of fuzz task's bad build count "
                 '(grouped by job type)'),
    field_spec=[
        monitor.StringField('job'),
        monitor.BooleanField('bad_build'),
    ])

FUZZER_KNOWN_CRASH_COUNT = monitor.CounterMetric(
    'task/fuzz/fuzzer/known_crash_count',
    description=('Count of fuzz task\'s known crash count '
                 '(grouped by fuzzer)'),
    field_spec=[
        monitor.StringField('fuzzer'),
    ])

FUZZER_NEW_CRASH_COUNT = monitor.CounterMetric(
    'task/fuzz/fuzzer/new_crash_count',
    description=('Count of fuzz task\'s new crash count '
                 '(grouped by fuzzer)'),
    field_spec=[
        monitor.StringField('fuzzer'),
    ])

JOB_KNOWN_CRASH_COUNT = monitor.CounterMetric(
    'task/fuzz/job/known_crash_count',
    description=('Count of fuzz task\'s known crash count '
                 '(grouped by job)'),
    field_spec=[
        monitor.StringField('job'),
    ])

JOB_NEW_CRASH_COUNT = monitor.CounterMetric(
    'task/fuzz/job/new_crash_count',
    description=('Count of fuzz task\'s new crash count '
                 '(grouped by job)'),
    field_spec=[
        monitor.StringField('job'),
    ])

FUZZER_RETURN_CODE_COUNT = monitor.CounterMetric(
    'task/fuzz/fuzzer/return_code_count',
    description=("Count of fuzz task's return codes "
                 '(grouped by fuzzer)'),
    field_spec=[
        monitor.StringField('fuzzer'),
        monitor.IntegerField('return_code'),
    ],
)

FUZZER_TOTAL_FUZZ_TIME = monitor.CounterMetric(
    'task/fuzz/fuzzer/total_time',
    description=('The total fuzz time in seconds '
                 '(grouped by fuzzer)'),
    field_spec=[
        monitor.StringField('fuzzer'),
        monitor.BooleanField('timeout'),
    ],
)

JOB_TOTAL_FUZZ_TIME = monitor.CounterMetric(
    'task/fuzz/job/total_time',
    description=('The total fuzz time in seconds '
                 '(grouped by job)'),
    field_spec=[
        monitor.StringField('job'),
        monitor.BooleanField('timeout'),
    ],
)

FUZZER_TESTCASE_COUNT_RATIO = monitor.CumulativeDistributionMetric(
    'task/fuzz/fuzzer/testcase_count_ratio',
    bucketer=monitor.FixedWidthBucketer(width=0.05, num_finite_buckets=20),
    description=('Distribution of fuzz task\'s generated testcase '
                 'counts divided by expected testcase counts '
                 '(grouped by fuzzers)'),
    field_spec=[
        monitor.StringField('fuzzer'),
    ],
)

# Global log_error count.
LOG_ERROR_COUNT = monitor.CounterMetric(
    'errors/count',
    description='Error count.',
    field_spec=[
        monitor.StringField('task_name'),
    ])

# Untrusted host metrics.
HOST_INCONSISTENT_COUNT = monitor.CounterMetric(
    'untrusted_runner/host/inconsistent_count',
    description='Inconsistent worker state count.',
    field_spec=None)

HOST_ERROR_COUNT = monitor.CounterMetric(
    'untrusted_runner/host/error_count',
    description='Error count.',
    field_spec=[
        monitor.IntegerField('return_code'),
    ])

TRY_COUNT = monitor.CounterMetric(
    'utils/retry/count',
    description='Success/Failure count when utils.retry fails',
    field_spec=[
        monitor.StringField('function'),
        monitor.BooleanField('is_succeeded'),
    ])

BOT_COUNT = monitor.GaugeMetric(
    'bot_count',
    description='Bot count',
    field_spec=[monitor.StringField('revision')])

TASK_COUNT = monitor.CounterMetric(
    'task/count',
    description='The number of started tasks',
    field_spec=[
        monitor.StringField('task'),
        monitor.StringField('job'),
    ])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""shell tests."""
import os
import sys
import unittest

import mock
from pyfakefs import fake_filesystem_unittest

from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class RemoveEmptyFilesTest(fake_filesystem_unittest.TestCase):
  """Tests for remove_empty_files."""

  def setUp(self):
    # FIXME: Add support for Windows.
    if not environment.is_posix():
      self.skipTest('Process tests are only applicable for posix platforms.')

    test_utils.set_up_pyfakefs(self)

  def test_remove(self):
    """Test remove."""
    self.fs.create_file('/test/aa/bb.txt', contents='s')
    self.fs.create_file('/test/aa/cc.txt', contents='')
    self.fs.create_file('/test/aa/aa/dd.txt', contents='s')
    self.fs.create_file('/test/aa/aa/aa.txt', contents='')

    shell.remove_empty_files('/test')

    self.assertTrue(os.path.exists('/test/aa/bb.txt'))
    self.assertTrue(os.path.exists('/test/aa/aa/dd.txt'))
    self.assertFalse(os.path.exists('/test/aa/cc.txt'))
    self.assertFalse(os.path.exists('/test/aa/aa/aa.txt'))

  def test_ignore_file(self):
    self.fs.create_file('/test/aa/cc.txt', contents='')
    shell.remove_empty_files('/test/aa/cc.txt')
    self.assertTrue(os.path.exists('/test/aa/cc.txt'))

  @mock.patch('os.remove', autospec=True)
  def test_exception(self, mock_remove):
    # bypass pyfakefs's os.remove.
    os.remove = mock_remove
    mock_remove.side_effect = OSError()

    self.fs.create_file('/test/aa/cc.txt', contents='')
    shell.remove_empty_files('/test')
    self.assertTrue(os.path.exists('/test/aa/cc.txt'))


class RemoveDirectoryTest(unittest.TestCase):
  """Tests for remove_directory."""

  def setUp(self):
    test_helpers.patch(self, [
        'os.chmod',
        'os.makedirs',
        'os.path.exists',
        'os.system',
        'clusterfuzz._internal.system.environment.platform',
        'clusterfuzz._internal.metrics.logs.log_error',
        'clusterfuzz._internal.metrics.logs.log_warn',
        'shutil.rmtree',
    ])

  def _test_remove_os_specific(self, platform, recreate, raise_makedirs_error):
    """Helper for testing removing dir with os-specific command."""
    self.mock.platform.return_value = platform
    self.mock.exists.side_effect = [True, False, False]
    if raise_makedirs_error:
      self.mock.makedirs.side_effect = OSError()

    result = shell.remove_directory('dir', recreate=recreate)

    if recreate:
      self.assertEqual(not raise_makedirs_error, result)
    else:
      self.assertTrue(result)

    self.mock.rmtree.assert_has_calls([])
    if recreate:
      self.mock.makedirs.assert_has_calls([mock.call('dir')])
    else:
      self.mock.makedirs.assert_has_calls([])

  def test_remove_os_specific_windows(self):
    """Test remove with os-specific command on windows."""
    self._test_remove_os_specific('WINDOWS', True, False)
    self.mock.system.assert_has_calls([mock.call('rd /s /q "dir" > nul 2>&1')])

  def test_remove_os_specific_non_windows(self):
    """Test remove with os-specific command on non-windows."""
    self._test_remove_os_specific('LINUX', True, False)
    self.mock.system.assert_has_calls(
        [mock.call('rm -rf "dir" > /dev/null 2>&1')])

  def test_remove_without_recreate(self):
    """Test remove without recreate."""
    self._test_remove_os_specific('LINUX', False, True)

  def test_remove_with_makedirs_error(self):
    """Test remove when makedirs errors."""
    self._test_remove_os_specific('LINUX', True, True)

  def test_remove_shutil_success(self):
    """Test remove with shutil."""
    self.mock.exists.side_effect = [True, True, False]
    self.assertTrue(shell.remove_directory('dir'))
    self.mock.system.assert_has_calls(
        [mock.call('rm -rf "dir" > /dev/null 2>&1')])
    self.mock.rmtree.assert_has_calls([mock.call('dir', onerror=mock.ANY)])

  def test_remove_shutil_failure(self):
    """Test remove with shutil but fails."""
    self.mock.exists.side_effect = [True, True, True]
    self.assertFalse(shell.remove_directory('dir'))
    self.mock.log_error.assert_has_calls(
        [mock.call('Failed to clear directory dir.')])
    self.assertEqual(0, self.mock.log_warn.call_count)
    self.mock.system.assert_has_calls(
        [mock.call('rm -rf "dir" > /dev/null 2>&1')])
    self.mock.rmtree.assert_has_calls([mock.call('dir', onerror=mock.ANY)])

  def test_remove_shutil_failure_ignore_errors(self):
    self.mock.exists.side_effect = [True, True, True]
    self.assertFalse(shell.remove_directory('dir', ignore_errors=True))
    self.mock.log_warn.assert_has_calls(
        [mock.call('Failed to clear directory dir.')])
    self.assertEqual(0, self.mock.log_error.call_count)
    self.mock.system.assert_has_calls(
        [mock.call('rm -rf "dir" > /dev/null 2>&1')])
    self.mock.rmtree.assert_has_calls([mock.call('dir', onerror=mock.ANY)])

  def test_remove_shutil_onerror(self):
    """Test shutil invoking onerror."""
    self.mock.exists.side_effect = [True, True, False]
    self.assertTrue(shell.remove_directory('dir'))
    self.mock.system.assert_has_calls(
        [mock.call('rm -rf "dir" > /dev/null 2>&1')])
    self.mock.rmtree.assert_has_calls([mock.call('dir', onerror=mock.ANY)])

    onerror = self.mock.rmtree.call_args[1]['onerror']
    fake_fn = mock.MagicMock()
    fake_fn.side_effect = OSError()

    onerror(fake_fn, 'dir/child', ImportError())

    self.mock.chmod.assert_has_calls([mock.call('dir/child', 0o750)])
    fake_fn.assert_has_calls([mock.call('dir/child')])


class GetDirectoryFileCount(fake_filesystem_unittest.TestCase):
  """Tests for get_directory_file_count."""

  def setUp(self):
    test_utils.set_up_pyfakefs(self)

  def test(self):
    """Test get_directory_file_count."""
    self.fs.create_file('/test/aa/bb.txt', contents='abc')
    self.fs.create_file('/test/aa/cc.txt', contents='def')
    self.fs.create_file('/test/aa/aa/aa.txt', contents='ghi')
    self.fs.create_file('/test/aa/aa/dd.txt', contents='t')

    self.assertEqual(shell.get_directory_file_count('/test/aa'), 4)


class GetDirectorySizeTest(fake_filesystem_unittest.TestCase):
  """Tests for get_directory_size."""

  def setUp(self):
    test_utils.set_up_pyfakefs(self)

  def test(self):
    """Test get_directory_size."""
    self.fs.create_file('/test/aa/bb.txt', contents='abc')
    self.fs.create_file('/test/aa/cc.txt', contents='def')
    self.fs.create_file('/test/aa/aa/aa.txt', contents='ghi')
    self.fs.create_file('/test/aa/aa/dd.txt', contents='t')

    self.assertEqual(shell.get_directory_size('/test/aa'), 10)


class WhichTest(fake_filesystem_unittest.TestCase):
  """Tests for which (shutil.which)."""

  def setUp(self):
    # FIXME: Add support for Windows.
    if not environment.is_posix():
      self.skipTest('Which test is only supported on posix platforms.')

  def test(self):
    self.assertTrue(shell.which('ls') in ['/bin/ls', '/usr/bin/ls'])


class ClearSystemTempDirectoryTest(fake_filesystem_unittest.TestCase):
  """Tests for clear_system_temp_directory."""

  def setUp(self):
    test_helpers.patch(self, [
        'tempfile.gettempdir',
    ])
    self.mock.gettempdir.return_value = '/tmp'

    test_utils.set_up_pyfakefs(self)

  def test(self):
    """Test clear_system_temp_directory works as expected."""
    self.fs.create_file('/tmp/aa/bb.txt', contents='abc')
    self.fs.create_file('/tmp/cc/dd/ee.txt', contents='def')
    self.fs.create_dir('/tmp/ff/gg')
    self.fs.create_dir('/tmp/hh')
    self.fs.create_dir('/unrelated')
    self.fs.create_file('/unrelated/zz.txt', contents='zzz')
    os.symlink('/unrelated/zz.txt', '/tmp/hh/gg.txt')
    os.symlink('/unrelated', '/tmp/ii')

    shell.clear_system_temp_directory()

    self.assertTrue(os.path.exists('/tmp'))
    self.assertTrue(os.path.exists('/unrelated'))
    self.assertEqual(shell.get_directory_file_count('/tmp'), 0)
    self.assertEqual(shell.get_directory_file_count('/unrelated'), 1)
    self.assertFalse(os.path.exists('/tmp/aa/bb.txt'))
    self.assertFalse(os.path.exists('/tmp/cc/dd/ee.txt'))
    self.assertFalse(os.path.exists('/tmp/ff/gg'))
    self.assertFalse(os.path.exists('/tmp/hh'))


class GetExecuteCommand(unittest.TestCase):
  """Test that the correct commands to run files are returned."""

  def call_and_assert_helper(self, expected_command, file_to_execute):
    """Call get_execute_command on |file_to_execute| and assert result equal to
    |expected_command|."""
    self.assertEqual(expected_command,
                     shell.get_execute_command(file_to_execute))

  def test_standard_script(self):
    """Test correct command returned for python script."""
    script_name = 'script.py'
    expected_command = sys.executable + ' ' + script_name
    self.call_and_assert_helper(expected_command, script_name)

  def test_java(self):
    """Test correct launch command returned for Java class."""
    script_name = 'javaclassfile.class'
    expected_command = 'java javaclassfile'
    self.call_and_assert_helper(expected_command, script_name)

  def test_binary(self):
    """Test correct launch command returned for a binary (executable) file."""
    executable_name = 'executable'
    self.call_and_assert_helper(executable_name, executable_name)

    executable_name += '.exe'
    self.call_and_assert_helper(executable_name, executable_name)


class GetInterpreter(object):
  """Test that the correct interpreters to execute a file are returned."""

  def get_interpreted_file_test(self):
    """Test correct interpreter is returned for a file that needs one."""
    self.assertEqual('python', shell.get_interpreter('run.py'))

  def get_non_interpreter_file_test(self):
    """Test that None is returned for a file that doesn't need one. We don't
    want empty string since this is easier to than None. """
    self.assertIsNone(shell.get_interpreter('executable'))
